"""
Contains objects used throughout puddletag
"""

import itertools
import json
import logging
import os
import re
import sys
import time
from bisect import bisect_left, insort_left  # for unique function.
from collections import defaultdict
from copy import copy
from functools import partial
from glob import glob
from io import StringIO
from itertools import groupby  # for unique function.
from typing import Any, Callable, Generator, List, Optional, Tuple, Union

from PyQt5.QtCore import QBuffer, QByteArray, QCollator, QCollatorSortKey, QDir, QLocale, QObject, QRectF, QSettings, \
    QSize, QThread, QTimer, Qt, pyqtSignal
from PyQt5.QtCore import QFile, QIODevice
from PyQt5.QtGui import QIcon, QBrush, QPixmap, QImage, \
    QKeySequence
from PyQt5.QtSvg import QGraphicsSvgItem, QSvgRenderer
from PyQt5.QtWidgets import QAbstractItemView, QAction, QApplication, QComboBox, QDialog, QDialogButtonBox, \
    QDockWidget, QFileDialog, QFrame, QGraphicsPixmapItem, QGraphicsScene, QGraphicsView, QGridLayout, QHBoxLayout, \
    QHeaderView, QLabel, QLayout, QLineEdit, QListWidget, QMenu, QMessageBox, QProgressBar, QPushButton, QSizePolicy, \
    QTextEdit, QToolButton, QVBoxLayout, QWidget
from configobj import ConfigObjError

from . import audioinfo
from .audioinfo import (IMAGETYPES, DESCRIPTION, DATA, IMAGETYPE, DEFAULT_COVER,
                        INFOTAGS, get_mime)
from .constants import ACTIONDIR, SAVEDIR, CONFIGDIR
from .translations import translate

path = os.path

# Parameters for string distance function.
# Words that can be moved to the end of a string using a comma.
SD_END_WORDS = ['the', 'a', 'an']
# Reduced weights for certain portions of the string.
SD_PATTERNS = [
    (r'^the ', 0.1),
    (r'[\[\(]?(ep|single)[\]\)]?', 0.0),
    (r'[\[\(]?(featuring|feat|ft)[\. :].+', 0.1),
    (r'\(.*?\)', 0.3),
    (r'\[.*?\]', 0.3),
    (r'(, )?(pt\.|part) .+', 0.2),
]

mod_keys = {
    Qt.KeyboardModifier.ShiftModifier: 'Shift',
    Qt.KeyboardModifier.MetaModifier: 'Meta',
    Qt.KeyboardModifier.AltModifier: 'Alt',
    Qt.KeyboardModifier.ControlModifier: 'Ctrl',
    Qt.KeyboardModifier.NoModifier: '',
    Qt.KeyboardModifier.KeypadModifier: '',
    Qt.KeyboardModifier.GroupSwitchModifier: '', }


def keycmp(modifier):
    if modifier == Qt.Modifier.CTRL:
        return 4
    elif modifier == Qt.Modifier.SHIFT:
        return 3
    elif modifier == Qt.Modifier.ALT:
        return 2
    elif modifier == Qt.Modifier.META:
        return 1
    else:
        return 0


modifiers = {}
for i in range(1, len(mod_keys)):
    for keys in set(itertools.permutations(mod_keys, i)):
        mod = keys[0]
        for key in keys[1:]:
            mod = mod | key
        modifiers[int(mod)] = '+'.join(mod_keys[key] for key in sorted(keys, key=keycmp) if mod_keys[key])

mod_keys = set((Qt.Key.Key_Shift, Qt.Key.Key_Control, Qt.Key.Key_Meta, Qt.Key.Key_Alt))

imagetypes = [
    (translate('Cover Type', 'Other'), translate("Cover Type", 'O')),
    (translate('Cover Type', 'File Icon'), translate("Cover Type", 'I')),
    (translate('Cover Type', 'Other File Icon'), translate("Cover Type", 'OI')),
    (translate('Cover Type', 'Cover (front)'), translate("Cover Type", 'CF')),
    (translate('Cover Type', 'Cover (back)'), translate("Cover Type", 'CB')),
    (translate('Cover Type', 'Leaflet page'), translate("Cover Type", 'LF')),
    (translate('Cover Type', 'Media (e.g. label side of CD)'), translate("Cover Type", 'M')),
    (translate('Cover Type', 'Lead artist'), translate("Cover Type", 'LA')),
    (translate('Cover Type', 'Artist'), translate("Cover Type", 'A')),
    (translate('Cover Type', 'Conductor'), translate("Cover Type", 'C')),
    (translate('Cover Type', 'Band'), translate("Cover Type", 'B')),
    (translate("Cover Type", 'Composer'), translate("Cover Type", 'CP')),
    (translate("Cover Type", 'Lyricist'), translate("Cover Type", 'L')),
    (translate("Cover Type", 'Recording Location'), translate("Cover Type", 'RL')),
    (translate("Cover Type", 'During recording'), translate("Cover Type", 'DR')),
    (translate("Cover Type", 'During performance'), translate("Cover Type", 'DP')),
    (translate("Cover Type", 'Movie/video screen capture'), translate("Cover Type", 'MC')),
    (translate("Cover Type", 'A bright coloured fish'), translate("Cover Type", 'F')),
    (translate("Cover Type", 'Illustration'), translate("Cover Type", 'P')),
    (translate("Cover Type", 'Band/artist logotype'), translate("Cover Type", 'BL')),
    (translate("Cover Type", 'Publisher/Studio logotype'), translate("Cover Type", 'PL'))]


def trans_imagetypes():
    global imagetypes
    imagetypes = [
        (translate('Cover Type', 'Other'), translate("Cover Type", 'O')),
        (translate('Cover Type', 'File Icon'), translate("Cover Type", 'I')),
        (translate('Cover Type', 'Other File Icon'), translate("Cover Type", 'OI')),
        (translate('Cover Type', 'Cover (front)'), translate("Cover Type", 'CF')),
        (translate('Cover Type', 'Cover (back)'), translate("Cover Type", 'CB')),
        (translate('Cover Type', 'Leaflet page'), translate("Cover Type", 'LF')),
        (translate('Cover Type', 'Media (e.g. label side of CD)'), translate("Cover Type", 'M')),
        (translate('Cover Type', 'Lead artist'), translate("Cover Type", 'LA')),
        (translate('Cover Type', 'Artist'), translate("Cover Type", 'A')),
        (translate('Cover Type', 'Conductor'), translate("Cover Type", 'C')),
        (translate('Cover Type', 'Band'), translate("Cover Type", 'B')),
        (translate("Cover Type", 'Composer'), translate("Cover Type", 'CP')),
        (translate("Cover Type", 'Lyricist'), translate("Cover Type", 'L')),
        (translate("Cover Type", 'Recording Location'), translate("Cover Type", 'RL')),
        (translate("Cover Type", 'During recording'), translate("Cover Type", 'DR')),
        (translate("Cover Type", 'During performance'), translate("Cover Type", 'DP')),
        (translate("Cover Type", 'Movie/video screen capture'), translate("Cover Type", 'MC')),
        (translate("Cover Type", 'A bright coloured fish'), translate("Cover Type", 'F')),
        (translate("Cover Type", 'Illustration'), translate("Cover Type", 'P')),
        (translate("Cover Type", 'Band/artist logotype'), translate("Cover Type", 'BL')),
        (translate("Cover Type", 'Publisher/Studio logotype'), translate("Cover Type", 'PL'))]


class CoverButton(QPushButton):
    currentIndexChanged = pyqtSignal(int, name='currentIndexChanged')

    def __init__(self, *args):
        QPushButton.__init__(self, *args)
        menu = QMenu(self)

        def create(title, short, index):
            text = '[%s] %s' % (short, title)
            action = QAction(text, self)
            action.triggered.connect(lambda: self.setCurrentIndex(index))
            return action

        actions = [create(title, short, index) for index, (title, short)
                   in enumerate(imagetypes)]

        list(map(menu.addAction, actions))
        self.setMenu(menu)
        self.setCurrentIndex(3)

    def setCurrentIndex(self, index):
        try:
            self.setText(imagetypes[index][1])
        except IndexError:
            self.setText(imagetypes[DEFAULT_COVER][1])
        self.currentIndexChanged.emit(index)
        self._index = index

    def currentIndex(self):
        return self._index


class PuddleConfig(object):
    """Module that allows you to values from INI config files, similar to
    Qt's Settings module (Created it because PyQt5.4.3 has problems with
    saving and loading lists.

    Only two functions of interest:

    get -> load a key from a specified section
    set -> save a key section"""

    def __init__(self, filename=None):
        if not filename:
            filename = os.path.join(CONFIGDIR, 'puddletag.conf')
        self.filename = filename

        self.setSection = self.set
        self.load = self.get

    def get(self, section, key, default, getint=False):
        settings = self.data
        try:
            value = self.data[section][key]
        except KeyError:
            return default

        if isinstance(default, bool):
            if value is True or value == 'True':
                return True
            return False
        elif getint or isinstance(default, int):
            try:
                return int(value)
            except TypeError:
                return list(map(int, value))
        else:
            if value is None:
                return default
            return value

    def set(self, section=None, key=None, value=None):
        settings = self.data
        if isinstance(value, (str, bytes)):
            value = str(value)
        if section in self.data:
            settings[section][key] = value
        else:
            settings[section] = {}
            settings[section][key] = value
        self.save()

    def reload(self):
        self.data = defaultdict(lambda: {})
        if os.path.exists(self.filename):
            try:
                with open(self.filename, 'r', encoding='utf-8') as config_file:
                    self.data.update(json.load(config_file))
            except json.JSONDecodeError as e:
                print(f'Error parsing config file {self.filename}: {e}')
            except Exception as e:
                print(f'Unexpected error while reading config file {self.filename}: {e}')

    def save(self):
        actions = self.data.get('puddleactions')
        filename = self.filename
        if not os.path.exists(filename):
            dirname = os.path.dirname(filename)
            try:
                os.makedirs(dirname)
            except:
                pass

        with open(filename, 'w') as fo:
            fo.write(json.dumps(dict(self.data), indent=2))


    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, filename):
        logging.debug(f'reading config file {filename}')
        self._filename = filename
        self.savedir = os.path.dirname(filename)
        self.reload()

    def sections(self):
        return list(self.data.keys())


def _getSettings():
    filename = os.path.join(CONFIGDIR, 'windowsizes')
    return QSettings(filename, QSettings.Format.IniFormat)


def savewinsize(name, dialog, settings=_getSettings()):
    settings.setValue(name, dialog.saveGeometry())


def winsettings(name, dialog, settings=_getSettings()):
    if settings.value(name):
        dialog.restoreGeometry(settings.value(name))
    cevent = dialog.closeEvent

    def closeEvent(self, event=None):
        savewinsize(name, dialog)
        if event is None:
            cevent(self)
        else:
            cevent(event)

    setattr(dialog, 'closeEvent', closeEvent)


# Next three functions from beets: http://code.google.com/p/beets

def _levenshtein(s1, s2):
    """A nice DP edit distance implementation from Wikibooks:
    http://en.wikibooks.org/wiki/Algorithm_implementation/Strings/
    Levenshtein_distance#Python
    """
    if len(s1) < len(s2):
        return _levenshtein(s2, s1)
    if not s1:
        return len(s2)

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1
            deletions = current_row[j] + 1
            substitutions = previous_row[j] + (c1 != c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row

    return previous_row[-1]


def _string_dist_basic(str1, str2):
    """Basic edit distance between two strings, ignoring
    non-alphanumeric characters and case. Normalized by string length.
    """
    str1 = re.sub(r'[^a-z0-9]', '', str1.lower())
    str2 = re.sub(r'[^a-z0-9]', '', str2.lower())
    if not str1 and not str2:
        return 0.0
    return _levenshtein(str1, str2) / float(max(len(str1), len(str2)))


def ratio(str1, str2):
    """Gives an "intuitive" edit distance between two strings. This is
    an edit distance, normalized by the string length, with a number of
    tweaks that reflect intuition about text.
    """
    str1 = str1.lower()
    str2 = str2.lower()

    # Don't penalize strings that move certain words to the end. For
    # example, "the something" should be considered equal to
    # "something, the".
    for word in SD_END_WORDS:
        if str1.endswith(', %s' % word):
            str1 = '%s %s' % (word, str1[:-len(word) - 2])
        if str2.endswith(', %s' % word):
            str2 = '%s %s' % (word, str2[:-len(word) - 2])

    # Change the weight for certain string portions matched by a set
    # of regular expressions. We gradually change the strings and build
    # up penalties associated with parts of the string that were
    # deleted.
    base_dist = _string_dist_basic(str1, str2)
    penalty = 0.0
    for pat, weight in SD_PATTERNS:
        # Get strings that drop the pattern.
        case_str1 = re.sub(pat, '', str1)
        case_str2 = re.sub(pat, '', str2)

        if case_str1 != str1 or case_str2 != str2:
            # If the pattern was present (i.e., it is deleted in the
            # the current case), recalculate the distances for the
            # modified strings.
            case_dist = _string_dist_basic(case_str1, case_str2)
            case_delta = max(0.0, base_dist - case_dist)
            if case_delta == 0.0:
                continue

            # Shift our baseline strings down (to avoid rematching the
            # same part of the string) and add a scaled distance
            # amount to the penalties.
            str1 = case_str1
            str2 = case_str2
            base_dist = case_dist
            penalty += weight * case_delta
    dist = base_dist + penalty

    return 1 - dist


dirlevels = lambda a: len(a.split('/'))


def removeslash(x):
    while x.endswith('/'):
        return removeslash(x[:-1])
    return x


def create_buddy(text, control, box=None):
    label = QLabel(text)
    label.setBuddy(control)

    if not box:
        box = QHBoxLayout()
    elif box is True:
        box = QVBoxLayout()
    box.addWidget(label)
    box.addWidget(control, 1)

    return box


def dircmp(a, b):
    """Compare function to sort directories via parent.
So that the child is renamed before parent, thereby not
giving Permission Denied errors."""
    a, b = removeslash(a), removeslash(b)
    if a == b:
        return 0
    elif a in b and (dirlevels(a) != dirlevels(b)):
        return 1
    elif b in a and (dirlevels(a) != dirlevels(b)):
        return -1
    elif len(a) > len(b):
        return 1
    elif len(b) > len(a):
        return -1
    elif len(b) == len(a):
        return 0


def dircmp1(a, b):
    """Like dircmp, but returns dirs as being in the same directory as equal."""
    a, b = removeslash(a), removeslash(b)
    if a == b or (dirlevels(a) == dirlevels(b)):
        return 0
    elif a in b:
        return 1
    elif b in a:
        return -1
    else:
        return 0


def issubfolder(parent, child, level=1):
    parent, child = removeslash(parent), removeslash(child)
    if isinstance(parent, str):
        sep = str(os.path.sep)
    else:
        sep = os.path.sep
    if level is not None:
        if child.startswith(parent + sep) and dirlevels(parent) + level == dirlevels(child):
            return True
        return False
    else:
        if child.startswith(parent + sep) and dirlevels(parent) < dirlevels(child):
            return True
        return False


HORIZONTAL = 1
VERTICAL = 0


def get_icon(name: Optional[str] = None, fallback: Optional[str] = None) -> QIcon:
    """Return the icon with the given name from the current icon theme.

    If the theme does not contain such icon, fallback to built-in png of
    the same name. The fallback file can be overriden by providing a filename
    as the second argument.
    """
    if not name and not fallback:
        return QIcon()

    fallback = fallback or f'{name}.png'
    return QIcon.fromTheme(name, QIcon(f'icons:{fallback}'))


def get_languages(dirs=None):
    files = []
    if dirs is not None:
        for d in dirs:
            files.extend(glob(os.path.join(d, "*.qm")))
    d = QDir('translations:./')
    if not d.isEmpty():
        files.extend([os.path.join('translations:./', t) for t in
                      map(str, d.entryList(['*.qm']))])

    ret = {}
    get_name = lambda s: os.path.splitext(os.path.basename(s))[0]
    for f in files:
        ts_name = get_name(f)
        if ts_name.startswith('puddletag_'):
            ret[ts_name[len('puddletag_'):]] = f
        else:
            ret[ts_name] = f
    return ret


def singleerror(parent, msg):
    QMessageBox.warning(parent, 'Error', msg)


def errormsg(parent, msg, maximum):
    """Shows a messagebox containing an error message indicating that
    writing to filename has failed and asks the user to continue, stop,
    or continue without interruption.

    error is the error that caused the disruption.
    single is the number of files that are being written. If it is 1, then
    just a warningMessage is shown.

    Returns:
        True if yes to all.
        False if No.
        None if just yes."""
    if maximum > 1:
        mb = QMessageBox(QMessageBox.Icon.Warning, translate("Defaults", 'Error'),
                         msg + translate("Defaults", "<br /> Do you want to continue?"),
                         QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No | QMessageBox.StandardButton.YesToAll,
                         parent)
        mb.setDefaultButton(QMessageBox.StandardButton.Yes)
        mb.setEscapeButton(QMessageBox.StandardButton.No)
        ret = mb.exec_()
        if ret == QMessageBox.StandardButton.No:
            return False
        elif ret == QMessageBox.StandardButton.YesToAll:
            return True
    else:
        singleerror(parent, msg)


def safe_name(name, chars=r'/\*?"|:', to=None):
    """Make a filename safe for use (remove some special chars)

    If any special chars are found they are replaced by to."""
    if not to:
        to = ""
    else:
        to = str(to)
    escaped = ""
    for ch in name:
        if ch not in chars:
            escaped = escaped + ch
        else:
            escaped = escaped + to
    if not escaped: return '""'
    return escaped


def unique(seq, stable=False):
    """unique(seq, stable=False): return a list of the elements in seq in arbitrary
    order, but without duplicates.
    If stable=True it keeps the original element order (using slower algorithms)."""
    # Developed from Tim Peters version:
    #   http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/52560

    # if uniqueDebug and len(str(seq))<50: print "Input:", seq # For debugging.

    # Special case of an empty s:
    if not seq: return []

    # if it's a set:
    if isinstance(seq, set): return list(seq)

    if stable:
        # Try with a set:
        seqSet = set()
        result = []
        try:
            for e in seq:
                if e not in seqSet:
                    result.append(e)
                    seqSet.add(e)
        except TypeError:
            pass  # move on to the next method
        else:
            # if uniqueDebug: print "Stable, set."
            return result

        # Since you can't hash all elements, use a bisection on sorted elements
        result = []
        sortedElem = []
        try:
            for elem in seq:
                pos = bisect_left(sortedElem, elem)
                if pos >= len(sortedElem) or sortedElem[pos] != elem:
                    insort_left(sortedElem, elem)
                    result.append(elem)
        except TypeError:
            pass  # Move on to the next method
        else:
            # if uniqueDebug: print "Stable, bisect."
            return result
    else:  # Not stable
        # Try using a set first, because it's the fastest and it usually works
        try:
            u = set(seq)
        except TypeError:
            pass  # move on to the next method
        else:
            # if uniqueDebug: print "Unstable, set."
            return list(u)

        # Elements can't be hashed, so bring equal items together with a sort and
        # remove them out in a single pass.
        try:
            t = sorted(seq)
        except TypeError:
            pass  # Move on to the next method
        else:
            # if uniqueDebug: print "Unstable, sorted."
            return [elem for elem, group in groupby(t)]

    # Brute force:
    result = []
    for elem in seq:
        if elem not in result:
            result.append(elem)
    # if uniqueDebug: print "Brute force (" + ("Unstable","Stable")[stable] + ")."
    return result


def natural_sort_key(s: Union[str, List[str]], case_insensitive=True) -> QCollatorSortKey:
    """Return a sort-key for natural sorting the given string.

    Case-insensitive means uppercase- and lowercase-characters are treated equally.
    Natural means sorting numbers by their numeric value, e.g. 100 comes after 99.
    It also takes the global user preferences into account (e.g. LC_COLLATE).
    """
    if isinstance(s, list):
        # Join all elements with ascii unit separator
        s = '\x1f'.join(s)

    locale = QLocale.system().collation() if case_insensitive else QLocale.c()
    collator = QCollator(locale)
    collator.setCaseSensitivity(Qt.CaseSensitivity.CaseSensitive)
    collator.setNumericMode(True)
    return collator.sortKey(s)


def dupes(l, method=None):
    if method is None:
        method = lambda a, b: int(a == b)
    l = [{'key': z, 'index': i} for i, z in enumerate(l)]
    chars = chars = r'/\*?;"|:\''
    strings = sorted([(safe_name(z['key'].lower(), chars, ''), z['index'])
                      for z in l if z['key'] is not None])
    try:
        last = strings[0][0]
    except IndexError:
        return []
    groups = [[0]]
    for z, i in strings[1:]:
        if z is not None:
            val = method(last, z)
            if val >= 0.85:
                groups[-1].append(i)
            else:
                last = z
                groups.append([i])
    return [z for z in groups if len(z) > 1]


def getfiles(files: Union[str, List[str]], subfolders: bool = False) -> Generator[str, None, None]:
    """For the given path(s), yield all the files.

    If path does not exist, ignore it.
    If path is a directory, yield all the files in that directory.
    If subfolders is True, also recurse into subdirectories and yield their files.
    """
    if not isinstance(files, list):
        files = [files]

    for file in files:
        if not os.path.exists(file):
            continue
        if not os.path.isdir(file):
            yield file
            continue

        for dirpath, dirnames, filenames in os.walk(file):
            for filename in filenames:
                yield os.path.join(dirpath, filename)
            if not subfolders:
                # don't recurse deeper
                dirnames.clear()


def gettags(files):
    return (gettag(audio) for audio in files)


def gettag(f):
    try:
        return audioinfo.Tag(f)
    except:
        logging.exception('Error loading file %s', f)
        return


def translate_filename_pattern(pat):
    """Translate a shell PATTERN to a regular expression.

    There is no way to quote meta-characters.
    """
    # from fnmatch.py with slight modification
    pat = pat.strip()
    i, n = 0, len(pat)
    res = ''
    while i < n:
        c = pat[i]
        i = i + 1
        if c == '*':
            res = res + '.*'
        elif c == '?':
            res = res + '.'
        elif c == '[':
            j = i
            if j < n and pat[j] == '!':
                j = j + 1
            if j < n and pat[j] == ']':
                j = j + 1
            while j < n and pat[j] != ']':
                j = j + 1
            if j >= n:
                res = res + '\\['
            else:
                stuff = pat[i:j].replace('\\', '\\\\')
                i = j + 1
                if stuff[0] == '!':
                    stuff = '^' + stuff[1:]
                elif stuff[0] == '^':
                    stuff = '\\' + stuff
                res = '%s[%s]' % (res, stuff)
        else:
            res = res + re.escape(c)
    # return res + '\Z(?ms)'
    return res + r'\Z'


def fnmatch(pattern, files, matchcase=False):
    regexp = '|'.join(map(translate_filename_pattern,
                          [z.strip() for z in pattern.split(';')]))
    if matchcase:
        match = re.compile(regexp).match
    else:
        match = re.compile(regexp, re.I).match
    return list(filter(match, files))


def gettaglist():
    cparser = PuddleConfig()
    filename = os.path.join(cparser.savedir, 'usertags')
    try:
        lines = sorted(set([z.strip()
                            for z in open(filename, 'rt').read().split('\n')]))
    except (IOError, OSError):
        lines = audioinfo.FIELDS[::]
    return lines


def settaglist(tags):
    cparser = PuddleConfig()
    filename = os.path.join(cparser.savedir, 'usertags')
    f = open(filename, 'w')
    text = '\n'.join(sorted([z for z in tags if not z.startswith('__')]))
    f.write(text)
    f.close()


def load_actions():
    from . import findfunc
    basename = os.path.basename

    funcs = {}
    cparser = PuddleConfig()
    set_value = partial(cparser.set, 'puddleactions')
    get_value = partial(cparser.get, 'puddleactions')

    firstrun = get_value('firstrun', True)
    set_value('firstrun', False)
    convert = get_value('convert', True)
    order = get_value('order', [])

    if convert:
        set_value('convert', False)
        findfunc.convert_actions(SAVEDIR, ACTIONDIR)
        if order:
            old_order = dict([(basename(z), i) for i, z in
                              enumerate(order)])
            files = glob(os.path.join(ACTIONDIR, '*.action'))
            order = {}
            for f in files:
                try:
                    order[old_order[basename(f)]] = f
                except KeyError:
                    pass
            order = [z[1] for z in sorted(order.items())]
            set_value('order', order)

    files = glob(os.path.join(ACTIONDIR, '*.action'))
    if firstrun and not files:
        filenames = ['data:./caseconversion.action', 'data:./standard.action']
        files = list(map(open_resourcefile, filenames))
        set_value('firstrun', False)

        for fileobj, filename in zip(files, filenames):
            filename = os.path.join(ACTIONDIR, filename[2:])
            f = open(filename, 'w')
            f.write(fileobj.read())
            f.close()
        files = glob(os.path.join(ACTIONDIR, '*.action'))

    files = [z for z in order if z in files] + \
            [z for z in files if z not in order]

    funcs = []
    for f in files:
        action = findfunc.load_macro_info(f)
        funcs.append([action[0], action[1], f])
    return funcs


def open_resourcefile(filename):
    f = QFile(filename)
    f.open(QIODevice.OpenModeFlag.ReadOnly)
    return StringIO(str(f.readAll().data(), encoding='utf-8'))


def progress(func: Callable[..., Generator[Optional[Tuple[str, int]], None, None]],
             pstring: str, maximum: int, threadfin: Optional[Callable[[], None]] = None) -> Callable[..., None]:
    """To be used for functions that need a threaded progressbar.

    Note that this function will only (and is meant to) work on dialogs.

    func is the function that will be run by the thread. It should yield None
    while successful. Otherwise it should yield an errormsg and the number
    of files (this'll be used when calling errormsg).

    pstring is the progress message. This is shown with the number of times
    func yielded a value. For instance, pstring = 'Loading... ', and maximum = 20
    will show 'Loading... 1 of 20', 'Loading... 2 of 20', etc.on the progress
    bar.

    maximum is the maximum value of the progessbar.

    threadfin is the function to run when the thread has finished. Usually
    for cleanup stuff.

    Note that the function returns a function that expects a parent for
    the progess window as the first argument. This with the rest of the arguments
    passed to the returned function are used when calling func (except in the
    case where only the parent argument is passed).
    """

    def s(*args):

        focused = QApplication.focusWidget()
        if focused:
            focusedpar = focused.parentWidget()
        else:
            focusedpar = None

        parent = args[0]

        if len(args) > 1:
            f = func(*args)
        else:
            f = func()

        if maximum == 1:
            errors = next(f)
            if errors and \
                    not isinstance(errors, (str, int)):
                errormsg(parent, errors[0], 1)
            if threadfin:
                threadfin()
            return
        elif maximum > 1:
            win = ProgressWin(parent, maximum, pstring)
            win.show()
        else:
            return

        parent.showmessage = True

        def threadfunc() -> None:
            i = 0
            err = False
            while not win.wasCanceled:
                try:
                    temp = next(f)
                    if isinstance(temp, str):
                        thread.message.emit(temp)
                    elif isinstance(temp, int):
                        thread.set_max.emit(temp)
                    elif temp is not None:
                        thread.error.emit(
                            temp[0], temp[1])
                        err = True
                        break
                    else:
                        thread.win.emit(i)
                except StopIteration:
                    break
                i += 1

            if not err:
                thread.win.emit(-1)

        def threadexit(*args) -> None:
            if args[0] == -1:
                win.close()
                win.destroy()
                QApplication.processEvents()
                if threadfin:
                    threadfin()
                if focusedpar is not None:
                    try:
                        focusedpar.setFocus()
                    except RuntimeError:
                        pass
                return
            elif isinstance(args[0], str):
                if parent.showmessage:
                    ret = errormsg(parent, args[0], maximum)
                    if ret is True:
                        parent.showmessage = False
                    elif ret is False:
                        thread.win.emit(-1)
                        return
                if not win.isVisible():
                    win.show()
                while thread.isRunning():
                    pass
                thread.start()
            win.setValue(win.value + 1)

        def set_message(msg: str) -> None:
            if msg != win.label.text():
                win.label.setText(msg)
                QApplication.processEvents()

        def set_max(value: int) -> None:
            win.pbar.setMaximum(value)

        thread = PuddleThread(threadfunc, parent)
        thread.win.connect(threadexit)
        thread.error.connect(threadexit)
        thread.message.connect(set_message)
        thread.set_max.connect(set_max)
        thread.start()

    return s


def timemethod(method):
    def f(*args, **kwargs):
        name = method.__name__
        t = time.time()
        ret = method(*args, **kwargs)
        print(name, time.time() - t)
        return ret

    return f


class HeaderSetting(QDialog):
    """A dialog that allows you to edit the header of a TagTable widget."""
    headerChanged = pyqtSignal([list, list], name='headerChanged')

    def __init__(self, tags=None, parent=None, showok=True, showedits=True):

        QDialog.__init__(self, parent)

        self.listbox = ListBox()
        self.tags = [list(z) for z in tags]
        self.listbox.addItems([z[0] for z in self.tags])

        self.vbox = QVBoxLayout()
        self.vboxgrid = QGridLayout()
        self.textname = QLineEdit()
        self.tag = QComboBox()
        self.tag.addItems(sorted(INFOTAGS) + gettaglist())
        self.tag.setEditable(True)
        self.buttonlist = ListButtons()
        self.buttonlist.editButton.setVisible(False)

        if showedits:
            self.vboxgrid.addWidget(QLabel(translate("Column Settings", "Title")), 0, 0)
            self.vboxgrid.addWidget(self.textname, 0, 1)
            self.vboxgrid.addWidget(QLabel(translate("Defaults", "Field")), 1, 0)
            self.vboxgrid.addWidget(self.tag, 1, 1)
            self.vboxgrid.addLayout(self.buttonlist, 2, 0)
        else:
            self.vboxgrid.addLayout(self.buttonlist, 1, 0)
        self.vboxgrid.setColumnStretch(0, 0)

        self.vbox.addLayout(self.vboxgrid)
        self.vbox.addStretch()

        self.grid = QGridLayout()
        self.grid.addWidget(self.listbox, 1, 0)
        self.grid.addLayout(self.vbox, 1, 1)
        self.grid.setColumnStretch(1, 1)
        self.grid.setColumnStretch(0, 2)

        self.listbox.currentItemChanged.connect(
            self.fillEdits)

        self.listbox.itemSelectionChanged.connect(self.enableEdits)

        self.okbuttons = OKCancel()
        if showok is True:
            self.grid.addLayout(self.okbuttons, 2, 0, 1, 2)
        self.setLayout(self.grid)

        self.okbuttons.ok.connect(self.okClicked)
        self.okbuttons.cancel.connect(self.close)
        self.textname.textChanged.connect(self.updateList)
        self.buttonlist.add.connect(self.add)
        self.buttonlist.moveup.connect(self.moveup)
        self.buttonlist.movedown.connect(self.movedown)
        self.buttonlist.remove.connect(self.remove)
        self.buttonlist.duplicate.connect(self.duplicate)

        self.listbox.setCurrentRow(0)

    def enableEdits(self):
        if len(self.listbox.selectedItems()) > 1:
            self.textname.setEnabled(False)
            self.tag.setEnabled(False)
            return
        self.textname.setEnabled(True)
        self.tag.setEnabled(True)

    def remove(self):
        if len(self.tags) == 1: return
        self.textname.textChanged.disconnect(self.updateList)
        self.listbox.currentItemChanged.disconnect(self.fillEdits)
        self.listbox.removeSelected(self.tags)
        row = self.listbox.currentRow()
        # self.listbox.clear()
        # self.listbox.addItems([z[0] for z in self.tags])

        if row == 0:
            self.listbox.setCurrentRow(0)
        elif row + 1 < self.listbox.count():
            self.listbox.setCurrentRow(row + 1)
        else:
            self.listbox.setCurrentRow(self.listbox.count() - 1)
        self.fillEdits(self.listbox.currentItem(), None)
        self.textname.textChanged.connect(self.updateList)
        self.listbox.currentItemChanged.connect(self.fillEdits)

    def moveup(self):
        self.listbox.moveUp(self.tags)

    def movedown(self):
        self.listbox.moveDown(self.tags)

    def updateList(self, text):
        self.listbox.currentItem().setText(text)

    def fillEdits(self, current, prev):
        row = self.listbox.row(prev)
        try:  # An error is raised if the last item has just been removed
            if row > -1:
                self.tags[row][0] = str(self.textname.text())
                self.tags[row][1] = str(self.tag.currentText())
        except IndexError:
            pass

        row = self.listbox.row(current)
        if row > -1:
            self.textname.setText(self.tags[row][0])
            self.tag.setEditText(self.tags[row][1])

    def okClicked(self):
        row = self.listbox.currentRow()
        if row > -1:
            self.tags[row][0] = str(self.textname.text())
            self.tags[row][1] = str(self.tag.currentText())
        self.headerChanged.emit([z for z in self.tags])
        self.close()

    def add(self):
        row = self.listbox.count()
        self.tags.append(["", ""])
        self.listbox.addItem("")
        self.listbox.clearSelection()
        self.listbox.setCurrentRow(row)
        self.textname.setFocus()

    def duplicate(self):
        row = self.listbox.currentRow()
        if row < 0:
            return
        tag = self.tags[row][::]
        self.tags.append(tag)
        self.listbox.addItem(tag[0])
        self.listbox.clearSelection()
        self.listbox.setCurrentRow(self.listbox.count() - 1)
        self.textname.setFocus()


class ListBox(QListWidget):
    """Puddletag's replacement of QListWidget, because
    removing, moving and deleting items in a listbox
    is done a lot.

    First the modifier methods.
    removeSelected, moveUp and moveDown each does as the
    name implies. See docstrings for more info.

    connectToListButtons -> connects removeSelected etc. to
    the respective buttons in a ListButtons object.

    Attributes:
    editButton -> Set this to a button or control which will be enabled only
    when a single item is selected.

    yourlist -> The list that will be used in removeSelected et al, if None
    is passed when calling the function.."""

    def __init__(self, parent=None):
        QListWidget.__init__(self, parent)
        self.yourlist = None
        self.editButton = None
        self.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)

    def items(self):
        return list(map(self.item, range(self.count())))

    def selectionChanged(self, selected, deselected):
        if self.editButton:
            if len(self.selectedItems()) == 1:
                self.editButton.setEnabled(True)
            else:
                self.editButton.setEnabled(False)
        QListWidget.selectionChanged(self, selected, deselected)

    def connectToListButtons(self, listbuttons, yourlist=None):
        """Connect the moveUp, moveDown and removeSelected to the
        moveup, movedown and remove signals of listbuttons and
        sets the editButton.

        yourlist is used a the argument in these functions if
        no other yourlist is passed."""
        self.editButton = listbuttons.editButton
        listbuttons.moveup.connect(self.moveUp)
        listbuttons.movedown.connect(self.moveDown)
        listbuttons.remove.connect(self.removeSelected)
        self.yourlist = yourlist

    def removeSelected(self, yourlist=None, rows=None):
        """Removes the currently selected items.
        If yourlist is not None, then the selected
        items are removed for yourlist also. Note, that
        the indexes of the items in yourlist and the listbox
        have to correspond.

        If you want to remove anything other than the selected,
        just set rows to a list of integers."""
        if not yourlist:
            yourlist = self.yourlist
        if rows:
            rows = sorted(rows)
        else:
            rows = sorted([self.row(item) for item in self.selectedItems()])

        for i in range(len(rows)):
            self.takeItem(rows[i])
            if yourlist:
                try:
                    del (yourlist[rows[i]])
                except (KeyError, IndexError):
                    "The list doesn't have enough items or something"
            rows = [z - 1 for z in rows]

    def moveUp(self, yourlist=None, rows=None):
        """Moves the currently selected items up one place.
        If yourlist is not None, then the indexes of yourlist
        are updated in tandem. Note, that
        the indexes of the items in yourlist and the listbox
        have to correspond."""
        if not rows:
            rows = [self.row(item) for item in self.selectedItems()]
        rows = sorted(rows)
        if not yourlist:
            yourlist = self.yourlist
        currentrow = self.currentRow() - 1
        if 0 in rows:
            return

        [item.setSelected(False) for item in self.selectedItems()]
        for i in range(len(rows)):
            row = rows[i]
            item = self.takeItem(row)
            self.insertItem(row - 1, item)
            if yourlist:
                temp = copy(yourlist[row - 1])
                yourlist[row - 1] = yourlist[row]
                yourlist[row] = temp
        [self.item(row - 1).setSelected(True) for row in rows]
        self.setCurrentRow(currentrow)

    def moveDown(self, yourlist=None, rows=None):
        """See moveup. It's exactly the opposite."""
        if rows is None:
            rows = [self.row(item) for item in self.selectedItems()]
        if self.count() - 1 in rows:
            return
        [item.setSelected(False) for item in self.selectedItems()]
        if not yourlist:
            yourlist = self.yourlist
        rows = sorted(rows)
        if len(rows) == 0:
            rows.append(0)
        lastindex = rows[0]
        groups = {lastindex: [lastindex]}
        lastrow = lastindex
        for row in rows[1:]:
            if row - 1 == lastindex:
                groups[lastrow].append(row)
            else:
                groups[row] = [row]
                lastrow = row
            lastindex = row

        for group in groups:
            item = self.takeItem(group + len(groups[group]))
            if yourlist:
                temp = copy(yourlist[group + len(groups[group])])
                for index in reversed(groups[group]):
                    yourlist[index + 1] = copy(yourlist[index])
                yourlist[group] = temp
            self.insertItem(group, item)

        [self.item(row + 1).setSelected(True) for row in rows]

    def selectedItems(self):
        return [item for item in map(self.item, range(self.count())) if item.isSelected()]


class ListButtons(QVBoxLayout):
    """A Layout that contains five buttons usually
    associated with listboxes. They are
    add, edit, movedown, moveup and remove.

    Each button, when clicked sends signal with the
    buttons name. e.g. add sends SIGNAL("add").

    You can find them all in the widgets attribute."""
    addSignal = pyqtSignal(name='add')
    removeSignal = pyqtSignal(name='remove')
    moveupSignal = pyqtSignal(name='moveup')
    movedownSignal = pyqtSignal(name='movedown')
    editSignal = pyqtSignal(name='edit')
    duplicateSignal = pyqtSignal(name='duplicate')

    def __init__(self, parent=None):
        QVBoxLayout.__init__(self, parent)
        self.addButton = QToolButton()
        self.addButton.setIcon(get_icon('list-add'))
        self.addButton.setToolTip(translate("List Buttons", 'Add'))
        self.removeButton = QToolButton()
        self.removeButton.setIcon(get_icon('list-remove'))
        self.removeButton.setToolTip(translate("List Buttons", 'Remove'))
        self.removeButton.setShortcut('Delete')
        self.moveupButton = QToolButton()
        self.moveupButton.setArrowType(Qt.ArrowType.UpArrow)
        self.moveupButton.setToolTip(translate("List Buttons", 'Move Up'))
        self.movedownButton = QToolButton()
        self.movedownButton.setArrowType(Qt.ArrowType.DownArrow)
        self.movedownButton.setToolTip(translate("List Buttons", 'Move Down'))
        self.editButton = QToolButton()
        self.editButton.setIcon(get_icon('document-edit'))
        self.editButton.setToolTip(translate("List Buttons", 'Edit'))
        self.duplicateButton = QToolButton()
        self.duplicateButton.setIcon(get_icon('edit-copy'))
        self.duplicateButton.setToolTip(translate("List Buttons", 'Duplicate'))
        self.copyButton = QToolButton()
        self.copyButton.setToolTip(translate("List Buttons", 'Copy to clipboard'))
        self.pasteButton = QToolButton()
        self.pasteButton.setToolTip(translate("List Buttons", 'Paste from clipboard'))

        self.widgets = [self.addButton, self.editButton, self.duplicateButton,
                        self.removeButton, self.moveupButton, self.movedownButton]
        [self.addWidget(widget) for widget in self.widgets]
        self.insertStretch(4)
        self.insertSpacing(4, 6)
        [z.setIconSize(QSize(16, 16)) for z in self.widgets]
        self.addStretch()

        self.addButton.clicked.connect(self.addClicked)
        self.removeButton.clicked.connect(self.removeClicked)
        self.moveupButton.clicked.connect(self.moveupClicked)
        self.movedownButton.clicked.connect(self.movedownClicked)
        self.editButton.clicked.connect(self.editClicked)
        self.duplicateButton.clicked.connect(self.duplicateClicked)

    def connectToWidget(self, widget, add=None, edit=None, remove=None,
                        moveup=None, movedown=None, duplicate=None):
        l = ['add', 'edit', 'remove']
        if moveup:
            l.append('moveup')
        if movedown:
            l.append('movedown')
        if duplicate:
            l.append('duplicate')
        connections = dict([(z, v) for z, v in zip(l,
                                                   [add, edit, remove, moveup, movedown,
                                                    duplicate]) if v])
        connect = lambda a: getattr(self, a).connect(
            connections[a] if a in connections else getattr(widget, a))
        list(map(connect, l))

    def addClicked(self):
        self.addSignal.emit()

    def setEnabled(self, value):
        [w.setEnabled(value) for w in self.widgets]
        super(ListButtons, self).setEnabled(value)

    def removeClicked(self):
        self.removeSignal.emit()

    def moveupClicked(self):
        self.moveupSignal.emit()

    def movedownClicked(self):
        self.movedownSignal.emit()

    def editClicked(self):
        self.editSignal.emit()

    def duplicateClicked(self):
        self.duplicateSignal.emit()


class MoveButtons(QWidget):
    indexChanged = pyqtSignal(int, name='indexChanged')

    def __init__(self, arrayname, index=0, orientation=HORIZONTAL, parent=None):
        QWidget.__init__(self, parent)
        self.next = QPushButton(translate("List Buttons", '&>>'))
        self.prev = QPushButton(translate("List Buttons", '&<<'))
        if orientation == VERTICAL:
            box = QVBoxLayout()
            box.addWidget(self.next, 0)
            box.addWidget(self.prev, 0)
        else:
            box = QHBoxLayout()
            box.addWidget(self.prev)
            box.addWidget(self.next)

        self.arrayname = arrayname

        self.setLayout(box)
        self.index = index
        self.next.clicked.connect(self.nextClicked)
        self.prev.clicked.connect(self.prevClicked)

    @property
    def index(self):
        return self._currentindex

    @index.setter
    def index(self, index):
        try:
            if index >= len(self.arrayname) or index < 0:
                return
            else:
                self._currentindex = index
                if self._currentindex >= len(self.arrayname) - 1:
                    self.next.setEnabled(False)
                else:
                    self.next.setEnabled(True)

                if self._currentindex <= 0:
                    self.prev.setEnabled(False)
                else:
                    self.prev.setEnabled(True)
        except TypeError:
            "Probably arrayname is None or something."
            self.prev.setEnabled(False)
            self.next.setEnabled(False)

        if (not self.prev.isEnabled()) and (not self.next.isEnabled()):
            self.prev.hide()
            self.next.hide()
        else:
            self.prev.show()
            self.next.show()

        self.indexChanged.emit(index)

    def nextClicked(self):
        self.index += 1

    def prevClicked(self):
        self.index -= 1

    def updateButtons(self):
        self.index = self.index


class OKCancel(QHBoxLayout):
    """Yes, I know about QDialogButtonBox, but I'm not using PyQt5.2 here."""
    ok = pyqtSignal(name='ok')
    cancel = pyqtSignal(name='cancel')

    def __init__(self, parent=None):
        QHBoxLayout.__init__(self, parent)
        # QDialogButtonBox.__init__(self, parent)

        # self.addStretch()
        dbox = QDialogButtonBox()

        self.okButton = dbox.addButton(QDialogButtonBox.StandardButton.Ok)
        self.cancelButton = dbox.addButton(QDialogButtonBox.StandardButton.Cancel)
        self.addStretch()
        self.addWidget(dbox)

        self.okButton.setText(translate('Defaults', 'OK'))
        self.cancelButton.setText(translate('Defaults', 'Cancel'))
        # self.cancelButton = QPushButton("&Cancel")
        # self.okButton.setDefault(True)

        # self.addWidget(self.okButton)
        # self.addWidget(self.cancelButton)

        self.okButton.clicked.connect(self.yes)
        self.cancelButton.clicked.connect(self.no)

    def yes(self):
        self.ok.emit()

    def no(self):
        self.cancel.emit()


class LongInfoMessage(QDialog):
    def __init__(self, title, question, html, parent=None):
        QDialog.__init__(self, parent)
        winsettings('infomessage', self)
        question = QLabel(question)

        text = QTextEdit()
        text.setReadOnly(True)
        # text.setWordWrapMode(QTextOption.WrapMode.NoWrap)
        text.setHtml(html)

        okcancel = OKCancel()

        okcancel.ok.connect(self._ok)
        okcancel.cancel.connect(self.close)

        vbox = QVBoxLayout()
        self.setWindowTitle(title)
        vbox.addWidget(question)
        vbox.addWidget(text)
        vbox.addLayout(okcancel)
        self.setLayout(vbox)

    def _ok(self):
        self.close()
        self.accept()


class ArtworkLabel(QGraphicsView):
    newImages = pyqtSignal(list, name='newImages')
    clicked = pyqtSignal(name='clicked')

    def __init__(self, *args, **kwargs):
        super(ArtworkLabel, self).__init__(*args, **kwargs)

        pal = self.palette()
        pal.setBrush(self.backgroundRole(), QBrush(pal.window()))
        self.setAutoFillBackground(True)
        self.setPalette(pal)

        self._svg = QGraphicsSvgItem()
        self._pixmap = QGraphicsPixmapItem()
        self._pixmap.setTransformationMode(Qt.TransformationMode.SmoothTransformation)
        self._scene = QGraphicsScene()
        self._scene.addItem(self._svg)
        self._scene.addItem(self._pixmap)
        self._shown_pixmap = None
        self.setScene(self._scene)
        self.setSceneRect(QRectF())

        self.setAcceptDrops(True)

    def dragEnterEvent(self, event):
        mime = event.mimeData()
        if mime.hasUrls():
            event.accept()
        else:
            event.ignore()

    def dragMoveEvent(self, event):
        if event.mimeData().hasUrls():
            event.accept()
        else:
            event.ignore()

    def dropEvent(self, event):
        mime = event.mimeData()
        if mime.hasUrls():
            filenames = [str(z.toString()) for z in mime.urls()]
            self.newImages.emit(filenames)
        super(ArtworkLabel, self).dropEvent(event)

    def mousePressEvent(self, event):
        super(ArtworkLabel, self).mousePressEvent(event)
        if event.buttons() == Qt.MouseButton.LeftButton:
            self.clicked.emit()

    def resizeEvent(self, event=None):
        if event is not None:
            super(ArtworkLabel, self).resizeEvent(event)
        if self._svg.isVisible():
            item = self._svg
        else:
            item = self._pixmap
        self.setSceneRect(item.boundingRect())
        self.fitInView(item, Qt.AspectRatioMode.KeepAspectRatio)

    def setPixmap(self, pixmap, data=None):
        if isinstance(pixmap, str):
            renderer = QSvgRenderer(QByteArray(bytes(pixmap, 'utf-8')), self._svg)
            self._svg.setSharedRenderer(renderer)
            self._pixmap.setVisible(False)
            self._svg.setVisible(True)
        else:
            self._data = data
            self._pixmap.setPixmap(pixmap)
            self._svg.setVisible(False)
            self._pixmap.setVisible(True)
        self.resizeEvent()


class PicWidget(QWidget):
    """A widget that shows a file's pictures.

    images is a list of mutagen.id3.APIC objects.
    It allows the user to edit, save and delete whichever
    picture the user wants, by right-clicking on it.

    In addition, there are buttons to browse through
    all the pictures.

    Some important attributes are:
    currentImage -> The index of the current image
    maxImage -> Shows the current image fullsized.
    setImages -> Guess
    addImage -> Guess again...but it also shows and open file dialog.
    removeImage -> Removes the current image.
    next and prevImage -> Moves to the next and previous image.
    saveToFile -> Save the current image to file.
    showbuttons -> If True, the >> and << buttons are always shown. If False,
                    they are shown depending on context."""
    imageChanged = pyqtSignal(name='imageChanged')

    def __init__(self, images=None, imagetags=None, parent=None,
                 readonly=None, buttons=False):
        """Initialises the widget.

        images -> A list of images as described in the classes docstring.
        parent -> Qt parent
        readonly -> indexes of images that are readonly. Can be changed by modifying
                    the readonly attribute.
        buttons -> If True, then the Add, Edit, etc. Buttons are shown.
                   If False, then these functions can be found by right clicking
                   on the picture."""

        self._contextFormat = translate('Artwork Context', "{}/{}")

        QWidget.__init__(self, parent)
        self.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.sizePolicy().setVerticalStretch(0)
        self.sizePolicy().setHorizontalStretch(3)

        self.lastfilename = '~'
        self.currentFile = None
        self.filePattern = 'folder.jpg'

        self.label = ArtworkLabel()
        self.label.setFrameStyle(QFrame.Shape.Box)

        self.label.setMinimumSize(200, 170)
        if buttons:
            self.label.setMaximumSize(200, 170)
        self._itags = []

        self.label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self.label.newImages.connect(
            lambda filenames: self.addImages(self.loadPics(*filenames)))

        self._image_size = QLabel()
        self._image_size.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        self._image_desc = QLineEdit(self)

        if (hasattr(self._image_desc, 'setPlaceholderText')):
            self._image_desc.setPlaceholderText(translate("Artwork", 'Enter a description'))
        else:
            self._image_desc.setText('')

        self._image_desc.setToolTip(
            translate("Artwork",
                      '<p>Enter a description for the current cover.</p>'
                      '<p>For ID3 tags the description has to be different for each '
                      "cover as per the ID3 spec. If they don't differ then spaces "
                      'are appended to the description when the tag is saved.</p>'))
        self._image_desc.textEdited.connect(self.setDescription)
        controls = QVBoxLayout()

        if buttons:
            dbox = QVBoxLayout()
            label = QLabel(translate("Artwork", '&Description'))
            label.setBuddy(self._image_desc)
            dbox.addWidget(label)
            dbox.addWidget(self._image_desc)
            controls.addLayout(dbox)
            self._image_type = QComboBox(self)
            self._image_type.addItems(IMAGETYPES)
            dbox = QVBoxLayout()
            label = QLabel(translate("Artwork", '&Type'))
            label.setBuddy(self._image_type)
            dbox.addWidget(label)
            dbox.addWidget(self._image_type)
            controls.addLayout(dbox)
        else:
            self._image_type = CoverButton(self)
            hbox = QHBoxLayout()
            hbox.addWidget(self._image_desc, 1)
            hbox.addWidget(self._image_type)
            controls.addLayout(hbox)
        self._image_type.setToolTip(
            translate("Artwork",
                      '<p>Select a cover type for the artwork.</p>'))
        self._image_type.currentIndexChanged.connect(self.setType)

        self.showbuttons = True

        if not readonly:
            readonly = []
        self.readonly = readonly

        self.next = QToolButton()
        self.next.setArrowType(Qt.ArrowType.RightArrow)
        self.prev = QToolButton()
        self.prev.setArrowType(Qt.ArrowType.LeftArrow)
        self.next.clicked.connect(self.nextImage)
        self.prev.clicked.connect(self.prevImage)

        self._contextlabel = QLabel()
        self._contextlabel.setVisible(False)
        if buttons:
            movebuttons = QHBoxLayout()
            movebuttons.addStretch()
            movebuttons.addWidget(self.prev)
            movebuttons.addWidget(self.next)
            movebuttons.addWidget(self._contextlabel)
            movebuttons.addStretch()
        else:
            self.next.setArrowType(Qt.ArrowType.UpArrow)
            self.prev.setArrowType(Qt.ArrowType.DownArrow)
            movebuttons = QVBoxLayout()
            movebuttons.addStretch()
            movebuttons.addWidget(self.next)
            movebuttons.addWidget(self.prev)
            movebuttons.addStretch()

        vbox = QVBoxLayout()
        v = QVBoxLayout()
        if buttons:
            v.addWidget(self.label)
            v.addWidget(self._image_size)
        else:
            v.addStretch()
            v.addWidget(self.label)
            v.addWidget(self._image_size)
            v.addStretch()

        h = QHBoxLayout()
        h.addStretch()
        h.addLayout(v)
        if not buttons:
            h.addLayout(movebuttons)
            context_box = QHBoxLayout()
            context_box.setAlignment(Qt.AlignmentFlag.AlignHCenter)
            context_box.addWidget(self._contextlabel)
            vbox.addLayout(context_box)
        h.addStretch()
        vbox.addLayout(h)

        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.addLayout(controls)
        if buttons:
            vbox.addLayout(movebuttons)
        vbox.addStretch()
        vbox.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.label.clicked.connect(self.maxImage)

        hbox = QHBoxLayout()
        hbox.addLayout(vbox)
        hbox.addStrut(12)
        hbox.setSizeConstraint(QLayout.SizeConstraint.SetMinAndMaxSize)
        self.setLayout(hbox)

        if buttons:
            listbuttons = ListButtons()
            listbuttons.duplicateButton.hide()
            self.addpic = listbuttons.addButton
            self.removepic = listbuttons.removeButton
            self.editpic = listbuttons.editButton
            self.savepic = QToolButton()
            self.savepic.setIcon(get_icon('document-save'))
            self.savepic.setIconSize(QSize(16, 16))
            self.copypic = listbuttons.copyButton
            self.pastepic = listbuttons.pasteButton
            listbuttons.insertWidget(3, self.savepic)
            listbuttons.moveupButton.hide()
            listbuttons.movedownButton.hide()
            signal = 'clicked'
            hbox.addLayout(listbuttons)

        else:
            self.label.setContextMenuPolicy(Qt.ContextMenuPolicy.ActionsContextMenu)
            self.savepic = QAction(translate("Artwork", "&Save cover to file"), self)
            self.label.addAction(self.savepic)

            self.addpic = QAction(translate("Artwork", "&Add cover"), self)
            self.label.addAction(self.addpic)

            self.copypic = QAction(translate("Artwork", "C&opy cover"), self)
            self.label.addAction(self.copypic)

            self.pastepic = QAction(translate("Artwork", "&Paste cover"), self)
            self.label.addAction(self.pastepic)

            self.removepic = QAction(translate("Artwork", "&Remove cover"), self)
            self.label.addAction(self.removepic)

            self.editpic = QAction(translate("Artwork", "&Change cover"), self)
            self.label.addAction(self.editpic)
            signal = 'triggered'

        getattr(self.addpic, signal).connect(self.addImage)
        getattr(self.removepic, signal).connect(self.removeImage)
        self.edit = partial(self.addImage, True)
        getattr(self.editpic, signal).connect(self.edit)
        getattr(self.savepic, signal).connect(self.saveToFile)
        getattr(self.copypic, signal).connect(self.copyImage)
        getattr(self.pastepic, signal).connect(self.pasteImage)

        # Put a listener on clipboard changes for checking if there's an image
        QApplication.clipboard().dataChanged.connect(self.clipboardChange)
        # Call it manually for the first time status
        self.clipboardChange()

        self.win = PicWin(parent=self)
        self._currentImage = -1

        if not images:
            images = []

        if not imagetags:
            imagetags = []

        self.setImages(images, imagetags)

        self._lastdata = None

    @property
    def context(self):
        return self._contextlabel.text()

    @context.setter
    def context(self, text):
        if not text:
            self._contextlabel.setVisible(False)
            self._contextlabel.setText('')
        else:
            self._contextlabel.setText(translate("Artwork Context", text))
            self._contextlabel.setVisible(True)

    def setDescription(self, text):
        '''Sets the description of the current image to the text in the
            description text box.'''
        self.images[self.currentImage]['description'] = str(text)
        self.imageChanged.emit()

    def setType(self, index):
        """Like setDescription, but for imagetype"""
        try:
            self.images[self.currentImage]['imagetype'] = index
            self.imageChanged.emit()
        except IndexError:
            pass

    def addImage(self, edit=False, filename=None):
        """Adds an image from the given filename to self.images.

        If a filename is not given, then an open file dialog is shown.
        If edit is True, then the current image is changed."""

        if not filename:
            default_fn = os.path.join(
                os.path.dirname(self.lastfilename), 'folder.jpg')
            selectedFile = QFileDialog.getOpenFileName(self,
                                                       translate("Artwork", 'Select Image...'), default_fn,
                                                       translate("Artwork", "JPEG & PNG Images (*.jpg *.jpeg *.png);;JPEG Images (*.jpg *.jpeg);;PNG Images (*.png);;All Files(*.*)"))
            filename = selectedFile[0]

        if not filename:
            return
        self.lastfilename = filename
        pic = self.loadPics(filename)
        if pic:
            pic = pic[0]
            if edit and self.images:
                self.images[self.currentImage].update(pic)
                self.currentImage = self.currentImage
            else:
                if not self.images:
                    self.setImages([pic])
                else:
                    self.images.append(pic)
                    self.currentImage = len(self.images) - 1
            self.imageChanged.emit()

    def pasteImage(self):
        image = QApplication.clipboard().image()
        if not image.isNull():
            ba = QByteArray()
            data = QBuffer(ba)
            data.open(QIODevice.OpenModeFlag.WriteOnly)
            # TODO: Don't transform to JPG
            image.save(data, "JPG")
            data = bytes(data.data())
            pic = {
                "data": data,
                "height": image.height(),
                "width": image.width(),
                "size": len(data),
                "mime": get_mime(data),
                "description": "",
                "imagetype": 3
            }
            self.addImages([pic])

    def copyImage(self):
        if self.images and self.currentImage:
            image = QImage()
            image.loadFromData(self.images[self.currentImage]["data"])
            QApplication.clipboard().setImage(image)

    def addImages(self, images):
        if not self._itags or not images:
            return
        if self.images:
            index = len(self.images)
            self.images.extend(images)
            self.currentImage = index
        else:
            self.setImages(images)
        self.imageChanged.emit()

    def close(self):
        self.win.close()
        QWidget.close(self)

    def enableButtons(self):
        """Enables or disables buttons depending on context.

        With < 1 image in self.images,
        they're hidden unless overidden by self.showbuttons."""
        if not self.images:
            self.next.setEnabled(False)
            self.prev.setEnabled(False)
        else:
            if self.currentImage >= len(self.images) - 1:
                self.next.setEnabled(False)
            else:
                self.next.setEnabled(True)
            if self.currentImage <= 0:
                self.prev.setEnabled(False)
            else:
                self.prev.setEnabled(True)

        if not self.showbuttons and not self.next.isEnabled() and not self.prev.isEnabled():
            self.next.hide()
            self.prev.hide()
        else:
            self.next.show()
            self.prev.show()

    @property
    def currentImage(self):
        """Get or set the index of the current image. If the index isn't valid
           then a blank image is loaded."""
        return self._currentImage

    @currentImage.setter
    def currentImage(self, num):
        while True:
            # A lot of files have corrupt picture data. I just want to
            # skip those and not have the user be any wiser.
            try:
                data = self.images[num]['data']
            except IndexError:
                self.setNone()
                return

            if isinstance(data, bytes) and data.startswith(b'<?xml'):
                image = data
                break
            elif isinstance(data, str) and data.startswith('<?xml'):
                image = data
                break
            else:
                image = QPixmap()
                if not image.loadFromData(data):
                    del (self.images[num])
                else:
                    break

        [action.setEnabled(True) for action in
         (self.editpic, self.savepic, self.removepic, self.copypic)]

        if hasattr(self, '_itags'):
            self.setImageTags(self._itags)
        if num in self.readonly:
            self.editpic.setEnabled(False)
            self.removepic.setEnabled(False)
            self.copypic.setEnabled(False)
            self._image_desc.setEnabled(False)
            self._image_type.setEnabled(False)
            self.savepic.setEnabled(False)
        if data != self._lastdata or self._lastdata is None:
            if isinstance(image, str):
                self.label.setPixmap(image, data)
                self.win.setImage(image)
                self.pixmap = None
            else:
                self.pixmap = image
                self.label.setPixmap(self.pixmap, data)
                self.win.setImage(self.pixmap)

        if isinstance(image, QPixmap):
            self._image_size.setText(str(image.width()) + "x" + str(image.height()))
        else:
            self._image_size.setText("")

        self._lastdata = data
        self._image_desc.blockSignals(True)
        desc = self.images[num].get('description',
                                    translate("Artwork", 'Enter a description'))
        self._image_desc.setText(desc)
        self._image_desc.blockSignals(False)
        self._image_type.blockSignals(True)
        try:
            self._image_type.setCurrentIndex(self.images[num]['imagetype'])
        except KeyError:
            self._image_type.setCurrentIndex(3)
        self._image_type.blockSignals(False)
        self._currentImage = num
        self.context = self._contextFormat.format(str(num + 1), str(len(self.images)))
        self.label.setFrameStyle(QFrame.Shape.NoFrame)
        self.enableButtons()
        # self.resizeEvent()

    def maxImage(self):
        """Shows a window with the picture fullsized."""
        if self.win.isVisible():
            self.win.hide()
        elif self.currentImage not in self.readonly:
            self.win = PicWin(self.pixmap, self)
            self.win.show()

    def nextImage(self):
        self.currentImage += 1

    def prevImage(self):
        self.currentImage -= 1

    def saveToFile(self):
        """Opens a dialog that allows the user to save,
        the image in the current file to disk."""
        from .functions import save_artwork

        if self.currentFile is not None and self.filePattern:
            tempfilename = save_artwork(self.currentFile,
                                        self.filePattern, self.currentFile, write=False)
            if not tempfilename:
                tempfilename = os.path.join(self.currentFile.dirpath,
                                            'folder.jpg')
        elif self.lastfilename:
            tempfilename = os.path.join(os.path.dirname(self.lastfilename),
                                        'folder.jpg')
        else:
            tempfilename = 'folder.jpg'
        if self.currentImage > -1:
            selectedFile = QFileDialog.getSaveFileName(
                self,
                translate("Artwork", 'Save artwork as...'),
                tempfilename,
                translate("Artwork", "JPEG Images (*.jpg);;PNG Images (*.png);;All Files(*.*)"))
            filename = selectedFile[0]
            if not filename:
                return
            if not self.pixmap.save(filename):
                QMessageBox.critical(self,
                                     translate('Defaults', "Error"),
                                     translate('Artwork', "Writing to <b>{}</b> failed.").format(filename)
                                     )

    def setNone(self):
        self.label.setFrameStyle(QFrame.Shape.Box)
        self.label.setPixmap(QPixmap())
        self._image_size.setText("")
        self.pixmap = None
        self.images = []
        self._image_desc.setEnabled(False)
        self._image_type.setEnabled(False)
        [action.setEnabled(False) for action in
         (self.editpic, self.savepic, self.removepic)]
        self.context = 'No Images'
        self._lastdata = None

    def setImages(self, images, imagetags=None, default=0):
        """Sets images. images are dictionaries as described in the class docstring."""
        if imagetags:
            self.setImageTags(imagetags)
        if images:
            self.images = images
            self.currentImage = default
        else:
            self.setNone()
        self.enableButtons()

    def removeImage(self):
        """Removes the current image."""
        if len(self.images) >= 1:
            del (self.images[self.currentImage])
            if self.currentImage >= len(self.images) - 1 and self.currentImage > 0:
                self.currentImage = len(self.images) - 1
            else:
                self.currentImage = self.currentImage
        self.imageChanged.emit()

    def loadPics(self, *filenames):
        """Loads pictures from the filenames.

        The filenames need to be passes as str arguments, one filename
        per argument. Lists and tuples need to be unpacked by the caller."""
        # I really need to sort out these circular references.
        from .tagsources import RetrievalError, urlopen
        images = []

        for filename in filenames:
            image = QImage()
            if filename.startswith(":/"):
                ba = QByteArray()
                data = QBuffer(ba)
                data.open(QIODevice.OpenModeFlag.WriteOnly)
                image.save(data, "JPG")
                data = str(data.data())
            else:
                try:
                    data = urlopen(filename)
                except (ValueError, RetrievalError):
                    try:
                        data = open(filename, 'rb').read()
                    except EnvironmentError:
                        continue

            if image.loadFromData(data):
                pic = {'data': data, 'height': image.height(),
                       'width': image.width(), 'size': len(data),
                       'mime': get_mime(data),
                       'description': "",
                       'imagetype': 3}
                images.append(pic)

        return images

    def picsFromData(self, *data):
        images = []
        for d in data:
            image = QImage().fromData(d)
            pic = {'data': d, 'height': image.height(),
                   'width': image.width(), 'size': len(data),
                   'mime': get_mime(d),
                   'description': "",
                   'imagetype': 3}
            images.append(pic)
        return images

    def setImageTags(self, itags):
        tags = {DESCRIPTION: self._image_desc.setEnabled,
                DATA: self.label.setEnabled,
                IMAGETYPE: self._image_type.setEnabled}
        self.enableButtons()
        if not itags:
            self.addpic.setEnabled(False)
        else:
            self.addpic.setEnabled(True)
        for z in itags:
            try:
                tags[z](True)
            except KeyError:
                pass

        others = [z for z in tags if z not in itags]
        if len(others) == len(tags):
            self.next.setEnabled(False)
            self.prev.setEnabled(False)
        for z in others:
            tags[z](False)
        self._itags = itags

    def clipboardChange(self):
        """Test if clipboard has a valid image, and enable menu according it."""
        image = QApplication.clipboard().image()
        self.pastepic.setEnabled(not image.isNull())


class PicWin(QDialog):
    """A windows that shows an image."""

    def __init__(self, pixmap=None, parent=None):
        """Loads the image specified in QPixmap pixmap.
        If picture is clicked, the window closes.

        If you don't want to load an image when the class
        is created, let pixmap = None and call setImage later."""
        QDialog.__init__(self, parent)
        self.setWindowTitle(QApplication.translate('Dialogs', 'Album Art'))
        self.label = ArtworkLabel()

        vbox = QVBoxLayout()
        vbox.setContentsMargins(0, 0, 0, 0)
        vbox.addWidget(self.label)
        self.setLayout(vbox)

        if pixmap is not None:
            self.setImage(pixmap)

        self.label.clicked.connect(self.close)

    def setImage(self, pixmap):
        maxsize = self.screen().availableGeometry().size()
        self.label.setPixmap(pixmap)
        if hasattr(pixmap, 'size'):
            size = pixmap.size()
            res = ": %sx%s" % (size.width(), size.height())
            self.setWindowTitle(self.windowTitle() + res)
            if size.height() < maxsize.height() and size.width() < maxsize.width():
                self.setMinimumSize(size)
                self.setMaximumSize(size)
            else:
                self.setMaximumSize(maxsize)
        else:
            self.setMaximumSize(maxsize)


class ProgressWin(QDialog):
    canceled = pyqtSignal(name='canceled')

    def __init__(self, parent: Optional[QObject] = None, maximum: int = 100, progresstext: str = '',
                 showcancel: bool = True) -> None:
        QDialog.__init__(self, parent)
        self._infunc = False
        self._cached = 0
        self.setModal(True)
        self.setWindowTitle(translate("Progress Dialog", "Please Wait..."))
        self._format = translate('Progress Dialog', "{}{} of {}...")

        self.ptext = progresstext

        self.pbar = QProgressBar(self)

        self.pbar.setRange(0, maximum)

        self.label = QLabel()
        self.label.setAlignment(Qt.AlignmentFlag.AlignHCenter)

        if maximum <= 0:
            self.pbar.setTextVisible(False)
            if not progresstext:
                self.label.setVisible(False)
            else:
                self.label.setText(progresstext)
            self.ptext = ''

        cancel = QPushButton(translate("Defaults", 'Cancel'))
        cbox = QHBoxLayout()
        cbox.addStretch()
        cbox.addWidget(cancel)
        if not showcancel:
            cancel.hide()

        vbox = QVBoxLayout()
        vbox.addWidget(self.label)
        vbox.addWidget(self.pbar)
        vbox.addLayout(cbox)
        self.setLayout(vbox)
        self.wasCanceled = False
        self.rejected.connect(self.cancel)
        cancel.clicked.connect(self.cancel)

        if maximum > 0:
            self.setValue(1)
        else:
            self._timer = QTimer(self)
            self._timer.setInterval(100)

            def update():
                self.setValue(self.pbar.value() + 1)

            self._timer.timeout.connect(update)

        if maximum <= 0:
            self._timer.start()

    def setValue(self, value: int) -> None:
        if self._infunc:
            return
        self._infunc = True
        if self.ptext:
            self.pbar.setTextVisible(False)
            self.label.setText(self._format.format(
                self.ptext, value, self.pbar.maximum()))
        self.pbar.setValue(value)
        self._infunc = False
        if self.pbar.maximum() and value >= self.pbar.maximum():
            self.close()

    def cancel(self) -> None:
        self.wasCanceled = True
        self.canceled.emit()
        self.close()

    def closeEvent(self, event):
        if hasattr(self, '_timer'):
            self._timer.stop()
        super(ProgressWin, self).closeEvent(event)

    @property
    def value(self) -> int:
        return self.pbar.value()


class PuddleCombo(QWidget):
    editTextChanged = pyqtSignal(str, name='editTextChanged')

    def __init__(self, name, default=None, parent=None):
        QWidget.__init__(self, parent)
        hbox = QHBoxLayout()
        hbox.setContentsMargins(0, 0, 0, 0)
        self.combo = QComboBox()
        self.combo.setSizeAdjustPolicy(QComboBox.SizeAdjustPolicy.AdjustToContents)

        self.remove = QToolButton()
        self.remove.setIcon(get_icon('list-remove'))
        self.remove.setToolTip(translate("Combo Box", 'Remove current item.'))
        self.remove.setIconSize(QSize(13, 13))
        self.remove.clicked.connect(self.removeCurrent)

        hbox.addWidget(self.combo)
        hbox.addWidget(self.remove)
        self.setLayout(hbox)

        self.combo.setEditable(True)

        self.setEditText = self.combo.setEditText
        self.currentText = self.combo.currentText

        self.name = name
        cparser = PuddleConfig()
        self.filename = os.path.join(os.path.dirname(cparser.filename), 'combos')
        if not default:
            default = []
        cparser.filename = self.filename
        items = cparser.load(self.name, 'values', default)
        newitems = []
        [newitems.append(z) for z in items if z not in newitems]
        self.combo.addItems(newitems)
        self.combo.editTextChanged.connect(
            self._editTextChanged)

    def load(self, name=None, default=None):
        if name:
            self.name = name
        if not default:
            default = []
        self.combo.clear()
        cparser = PuddleConfig(self.filename)
        self.combo.addItems(cparser.load(self.name, 'values', default))

    def save(self):
        values = [str(self.combo.itemText(index)) for index in range(self.combo.count())]
        values.append(str(self.combo.currentText()))
        cparser = PuddleConfig(self.filename)
        try:
            cparser.setSection(self.name, 'values', values)
        except ConfigObjError:
            pass

    def removeCurrent(self):
        self.combo.removeItem(self.combo.currentIndex())

    def _editTextChanged(self, text):
        self.editTextChanged.emit(text)

    def closeEvent(self, event):
        QWidget.closeEvent(self, event)

        self.save()


class PuddleDock(QDockWidget):
    """A normal QDockWidget that emits a 'visibilitychanged' signal
    when...uhm...it changes visibility."""
    _controls = {}
    visibilitychanged = pyqtSignal(bool, name='visibilitychanged')

    def __init__(self, title, control=None, parent=None, status=None):
        QDockWidget.__init__(self, translate("Dialogs", title), parent)
        self.title = title
        if control:
            control = control(status=status)
            self.setObjectName(title)
            self._control = control
            self._controls.update({title: control})
            self.setWidget(control)

    def setVisible(self, visible):
        QDockWidget.setVisible(self, visible)
        self.visibilitychanged.emit(visible)


class PuddleHeader(QHeaderView):
    def __init__(self, orientation=Qt.Orientation.Horizontal, parent=None):
        if parent:
            super(PuddleHeader, self).__init__(orientation, parent)
        else:
            super(PuddleHeader, self).__init__()

        self.setSortIndicatorShown(True)
        self.setSortIndicator(0, Qt.SortOrder.AscendingOrder)
        self.setSectionsMovable(True)
        self.setSectionsClickable(True)

    def getMenu(self, actions=None):
        model = self.model()

        def create_action(section):
            title = str(model.headerData(section, self.orientation()))
            action = QAction(title, self)
            action.setCheckable(True)

            def change_visibility(value):
                if value:
                    self.showSection(section)
                else:
                    self.hideSection(section)

            if self.isSectionHidden(section):
                action.setChecked(False)
            else:
                action.setChecked(True)
            action.toggled.connect(change_visibility)
            return action

        header_actions = [create_action(section)
                          for section in range(self.count())]

        menu = QMenu(self)
        if actions:
            [menu.addAction(a) for a in actions]
            menu.addSeperator()
        [menu.addAction(a) for a in header_actions]
        return menu

    def contextMenuEvent(self, event):
        menu = self.getMenu()
        menu.exec_(event.globalPos())


class PuddleStatus(object):
    _status = {}

    def __init__(self):
        object.__init__(self)

    def __setitem__(self, name, val):
        self._status[name] = val

    def __getitem__(self, name):
        x = self._status.get(name)
        if callable(x):
            return x()
        return x


class PuddleThread(QThread):
    """puddletag rudimentary threading.
    pass a command to run in another thread. The result
    is stored in retval."""
    threadfinished = pyqtSignal(object, name='threadfinished')
    statusChanged = pyqtSignal(str, name='statusChanged')
    enable_preview_mode = pyqtSignal(name='enable_preview_mode')
    setpreview = pyqtSignal(dict, name='setpreview')
    message = pyqtSignal(str, name='message')
    set_max = pyqtSignal(int, name='set_max')
    error = pyqtSignal([str, int], name='error')
    win = pyqtSignal(int, name='win')

    def __init__(self, command: Callable[[], Any], parent: Optional[QObject] = None) -> None:
        QThread.__init__(self, parent)
        self.finished.connect(self._finish)
        self.command = command
        self.retval = None

    def run(self) -> None:
        # print 'thread', self.command, time.time()
        try:
            self.retval = self.command()
        except StopIteration:
            self.retval = 'STOP'

    def _finish(self) -> None:
        if hasattr(self, 'retval'):
            self.threadfinished.emit(self.retval)
        else:
            self.threadfinished.emit(None)


class ShortcutEditor(QLineEdit):
    validityChanged = pyqtSignal(bool, name='validityChanged')

    def __init__(self, shortcuts=None, *args, **kwargs):
        QLineEdit.__init__(self, *args, **kwargs)
        winsettings('shortcutcapture', self)

        self.key = ""
        self.modifiers = {}
        self._valid = False
        if shortcuts is None:
            shortcuts = []
        self._shortcuts = shortcuts

    def clear(self):
        super(ShortcutEditor, self).clear()
        self.valid = False

    def keyPressEvent(self, event):

        text = ''

        if event.modifiers():
            text = modifiers[int(event.modifiers())]

        if event.key() not in mod_keys:
            if text:
                text += '+' + str(QKeySequence(event.key()).toString())
            else:
                text = str(QKeySequence(event.key()).toString())

            if text and text not in self._shortcuts:
                valid = True
            else:
                valid = False
        else:
            valid = False

        self.setText(text)
        self.valid = valid

    @property
    def valid(self):
        return self._valid

    @valid.setter
    def valid(self, value):
        self._valid = value
        self.validityChanged.emit(value)


if __name__ == '__main__':
    class MainWin(QDialog):
        def __init__(self, parent=None):
            QDialog.__init__(self, parent)
            self.combo = PuddleCombo('patterncombo',
                                     ['%artist% - $num(%track%, 2) - %title%', '%artist% - %title%', '%artist% - %album%', '%artist% - Track %track%', '%artist% - %title%', '%artist%'])

            hbox = QHBoxLayout()
            hbox.addWidget(self.combo)
            self.setLayout(hbox)

        def closeEvent(self, e):
            self.combo.save()
            QDialog.closeEvent(self, e)


    app = QApplication(sys.argv)
    widget = MainWin()
    widget.show()
    app.exec_()



import gzip
import json
import re
import socket
import time
import numbers
import urllib.error
import urllib.parse
import urllib.request
from copy import deepcopy
from io import StringIO

from puddlestuff.audioinfo import DATA, isempty
from puddlestuff.constants import CHECKBOX, COMBO, TEXT
from puddlestuff.tagsources import (
    find_id, write_log, RetrievalError, iri_to_uri, get_useragent)
from puddlestuff.util import translate

R_ID_DEFAULT = 'discogs_id'
R_ID = R_ID_DEFAULT
MASTER = 'master'
RELEASE = 'release'

SITE_MASTER_URL = 'https://www.discogs.com/master/'
SITE_RELEASE_URL = 'https://www.discogs.com/releases/'
API_RELEASE_URL = 'https://api.discogs.com/releases/'
API_MASTER_URL = 'https://api.discogs.com/masters/'
SITE_URL = 'https://www.discogs.com'

BASE_URL = 'https://api.discogs.com/%s'
SEARCH_URL = BASE_URL % 'database/search?type=release&q=%s&per_page=100'
RELEASE_URL = BASE_URL % 'releases/%s'
MASTER_URL = BASE_URL % 'masters/%s'

SMALLIMAGE = '#smallimage'
LARGEIMAGE = '#largeimage'
IMAGE_TYPES = [SMALLIMAGE, LARGEIMAGE]

IMAGEKEYS = {
    'SmallImage': SMALLIMAGE,
    'LargeImage': LARGEIMAGE}

ALBUM_KEYS = {
    'title': 'album',
    'uri': 'discogs_uri',
    'summary': 'discogs_summary',
    'released': 'year',
    'notes': 'discogs_notes',
    'id': '#r_id',
    'thumb': '#cover_url',
    'resource_url': '#discogs_url',
    'type': '#release_type',
    'master_url': "discogs_master_url"}

TRACK_KEYS = {
    'position': 'track',
    'duration': '__length',
    'notes': 'discogs_notes'}

RELEASE_PROPERTIES_TO_IGNORE = [
    'status', 'resource_url', 'tracklist', 'thumb', 'formats', 'artists',
    'extraartists', 'images', 'videos', 'master_id', 'labels', 'companies',
    'series', 'released_formatted', 'identifiers', 'sub_tracks']

__minimum_request_separation = 1  # second - no more than one request sent to discogs in this time frame
__reference_time = 0  # A reference time. The initial value ensure first request goes ahead


def convert_dict(d, keys=None):
    if keys is None:
        keys = TRACK_KEYS
    d = deepcopy(d)
    for key in keys:
        if key in d:
            d[keys[key]] = d[key]
            del (d[key])
    return d


def capture_dict_values(d):
    ret = {}
    for key, v in d.items():
        if key in RELEASE_PROPERTIES_TO_IGNORE or isempty(v) or isinstance(v, dict):
            continue
        elif isinstance(v, bytes):
            ret[key] = v.decode('utf8')
        elif isinstance(v, numbers.Number):
            ret[key] = str(v)
        else:
            ret[key] = v

    return ret


def parse_tracklist(tlist):
    tracks = []
    for t in tlist:
        if not t.get('duration') and not t.get('position'):
            continue
        title = t['title']
        people = []
        featuring = []
        for person in t.get('extraartists', {}):
            name = person['name']
            if person.get('join'):
                title = title + ' ' + person.get('join') + ' ' + name
            elif person.get('role', '') == 'Featuring':
                featuring.append(name)
            else:
                people.append("%s:%s" % (person['name'], person['role']))

        if featuring:
            title = title + ' featuring ' + ', '.join(featuring)

        info = convert_dict(t)
        artist = []
        a_len = len(t.get('artists', []))
        for i, a in enumerate(t.get('artists', [])):
            if a_len > 1 and a.get('join'):
                artist.append('%s %s ' % (a['name'], a['join']))
            else:
                if i < a_len - 1:
                    artist.append('%s & ' % a['name'])
                else:
                    artist.append(a['name'])
        info['artist'] = ''.join(artist).strip()
        info['title'] = title

        if people:
            info['involvedpeople_track'] = ';'.join(people)
        tracks.append(capture_dict_values(info))
    return tracks


def parse_album_json(data):
    """Parses the retrieved json for an album and get's the track listing."""

    info = {}

    formats = []
    for fmt in data.get('formats', []):
        desc = fmt.get('descriptions', fmt.get('name', ''))
        if not desc:
            continue
        if isinstance(desc, str):
            formats.append(desc)
        else:
            formats.extend(desc)

    if formats:
        info['format'] = list(set(formats))
    info['artist'] = [z['name'] for z in data.get('artists', [])]
    info['artist'] = " & ".join([_f for _f in info['artist'] if _f])
    info['involvedpeople_album'] = \
        ':'.join('%s;%s' % (z['name'], z['role'])
                 for z in data.get('extraartists', []))
    info['label'] = [z['name'] for z in data.get('labels', [])]
    info['catno'] = [_f for _f in (z.get('catno') for z in data.get('labels', [])) if _f]

    info['companies'] = ';'.join(
        '%s %s' % (z['entity_type_name'], z['name'])
        for z in data.get('companies', []))
    info['album'] = data['title']
    cleaned_data = convert_dict(capture_dict_values(data), ALBUM_KEYS)
    cleaned_data.update(capture_dict_values(convert_dict(info, ALBUM_KEYS)))
    info = cleaned_data

    if 'images' in data:
        images = \
            [(z.get('uri', ''), z.get('uri150', ''))
             for z in data['images'] if 'uri' in z or 'uri150' in z]
        info['#cover-url'] = images

    return info, parse_tracklist(data['tracklist'])

  
def retrieve_album(info, image=LARGEIMAGE, rls_type=None):
    """Retrieves album from the information in info.
    image must be either one of image_types or None.
    If None, no image is retrieved."""
    if isinstance(info, int):
        r_id = str(info)
        info = {}
        write_log(
            translate('Discogs', "Retrieving using Release ID: {}").format(r_id))
        rls_type = 'release'
    elif isinstance(info, str):
        r_id = info
        info = {}
        write_log(
            translate('Discogs', "Retrieving using Release ID: {}").format(r_id))
        rls_type = 'release'
    else:
        if rls_type is None and '#release_type' in info:
            rls_type = info['#release_type']
        r_id = info['#r_id']
        write_log(
            translate('Discogs', "Retrieving album {}").format(info['album']))

    site_url = SITE_MASTER_URL if rls_type == MASTER else SITE_RELEASE_URL
    site_url += r_id

    url = MASTER_URL % r_id if rls_type == MASTER else RELEASE_URL % r_id

    x = urlopen(url)
    ret = parse_album_json(json.loads(x))

    info = deepcopy(info)
    info.update(ret[0])

    if image in IMAGE_TYPES and '#cover-url' in info:
        data = []
        for large, small in info['#cover-url']:
            if image == LARGEIMAGE and large:
                write_log(
                    translate('Discogs', "Retrieving cover: {}").format(large))
                try:
                    data.append({DATA: urlopen(large)})
                except RetrievalError as e:
                    write_log(translate(
                        'Discogs', 'Error retrieving image:') + str(e))
            else:
                write_log(
                    translate('Discogs', "Retrieving cover: {}").format(small))
                try:
                    data.append({DATA: urlopen(small)})
                except RetrievalError as e:
                    write_log(translate(
                        'Discogs', 'Error retrieving image:') + str(e))
        if data:
            info.update({'__image': data})

    try:
        info['#extrainfo'] = translate(
            'Discogs', "{} at Discogs.com").format(info['album']), site_url
    except KeyError:
        pass
    return info, ret[1]


def urlopen(url):
    global __reference_time
    url = iri_to_uri(url)
    request = urllib.request.Request(url)
    request.add_header('Accept-Encoding', 'gzip')
    request.add_header('User-Agent', get_useragent())

    while (time.time() - __reference_time) < __minimum_request_separation:
        time.sleep(__minimum_request_separation)
    __reference_time = time.time()

    try:
        data = urllib.request.urlopen(request).read()
    except urllib.error.URLError as e:
        try:
            msg = '%s (%s)' % (e.reason.strerror, e.reason.errno)
        except AttributeError:
            msg = str(e)
        raise RetrievalError(msg)
    except socket.error as e:
        msg = '%s (%s)' % (e.strerror, e.errno)
        raise RetrievalError(msg)
    except EnvironmentError as e:
        msg = '%s (%s)' % (e.strerror, e.errno)
        raise RetrievalError(msg)

    try:
        data = gzip.decompress(data)
    except IOError:
        "Gzipped data not returned."

    return data


class Discogs(object):
    name = 'Discogs.com'
    group_by = ['album', 'artist']
    tooltip = translate("Discogs", """<p><b>Discogs only support searching by release id</b></p>
        <p>Enter the release id Eg. "1257896" to search.</p>""")

    def __init__(self):
        super(Discogs, self).__init__()
        self._getcover = True
        self.covertype = 1

        self.preferences = [
            [translate('Discogs', 'Retrieve Cover'), CHECKBOX, True],
            [translate("Discogs", 'Cover size to retrieve'), COMBO,
             [[translate("Discogs", 'Small'),
               translate("Discogs", 'Large')], 1]],
            [translate("Discogs", 'Field to use for discogs_id'), TEXT, R_ID],
            [translate("Discogs", 'API Key (Stored as plain-text.'
                                  'Leave empty to use default.)'), TEXT, ''],
        ]

    def keyword_search(self, text):

        try:
            r_id = int(text.strip())
        except (TypeError, ValueError):
            raise RetrievalError(translate(
                "Discogs", 'Discogs release id should be an integer.'))

        try:
            return [self.retrieve(r_id)]
        except Exception as e:
            raise RetrievalError(str(e))

    def search(self, album, artists):

        if len(artists) > 1:
            artist = 'Various Artists'
        else:
            artist = [z for z in artists][0]

        if hasattr(artists, 'values'):
            write_log(
                translate("Discogs", 'Checking tracks for Discogs Album ID.'))
            tracks = []
            [tracks.extend(z) for z in artists.values()]
            album_id = find_id(tracks, R_ID)
            if not album_id:
                write_log(
                    translate("Discogs", 'No Discogs ID found in tracks.'))
            else:
                write_log(
                    translate('Discogs', "Found Discogs ID: {}").format(album_id))
                return [self.retrieve(album_id)]

        return []

    def retrieve(self, info):
        if False and self._getcover:  # retrieving images is broken by Discogs
            return retrieve_album(info, self.covertype)
        else:
            return retrieve_album(info, None)

    def applyPrefs(self, args):
        self._getcover = args[0]
        self.covertype = IMAGE_TYPES[args[1]]
        if not self._getcover:
            self.covertype = None
        global R_ID
        if args[2]:
            R_ID = args[2]
        else:
            R_ID = R_ID_DEFAULT


info = Discogs

if __name__ == '__main__':
    k = Discogs()
    print(k.keyword_search(":r 911637"))



# Module for retrieving CDDB v1 data from CDDB servers via HTTP

# Written 17 Nov 1999 by Ben Gertzfield <che@debian.org>
# This work is released under the GNU GPL, version 2 or later.

# Release version 1.4
# CVS ID: $Id: CDDB.py,v 1.8 2003/08/31 23:18:43 che_fox Exp $

import os
import re
import socket
import urllib.error
import urllib.parse
import urllib.request

name = 'CDDB.py'
version = 1.4

if 'EMAIL' in os.environ:
    (default_user, hostname) = os.environ['EMAIL'].split('@')
else:
    default_user = os.geteuid() or os.environ['USER'] or 'user'
    hostname = socket.gethostname() or 'host'

# Use protocol version 5 to get DYEAR and DGENRE fields.
proto = 5
default_server = 'http://gnudb.gnudb.org/~cddb/cddb.cgi'


def query(track_info, server_url=default_server,
          user=default_user, host=hostname, client_name=name,
          client_version=version):
    disc_id = track_info[0]
    num_tracks = track_info[1]

    query_str = (('%08lx %d ') % (disc_id, num_tracks))

    for i in track_info[2:]:
        query_str = query_str + ('%d ' % i)

    query_str = urllib.parse.quote_plus(query_str.rstrip())

    url = "%s?cmd=cddb+query+%s&hello=%s+%s+%s+%s&proto=%i" % \
          (server_url, query_str, user, host, client_name,
           client_version, proto)

    response = urllib.request.urlopen(url)

    # Four elements in header: status, category, disc-id, title
    header = response.readline().decode().rstrip().split(' ', 3)

    header[0] = int(header[0])

    if header[0] == 200:  # OK
        result = {'category': header[1], 'disc_id': header[2], 'title':
            header[3]}

        return [header[0], result]

    elif header[0] == 211 or header[0] == 210:  # multiple matches
        result = []

        for line in response.readlines():
            line = line.decode().rstrip()

            if line == '.':  # end of matches
                break
                # otherwise:
                # split into 3 pieces, not 4
                # (thanks to bgp for the fix!)
            match = line.split(' ', 2)

            result.append({'category': match[0], 'disc_id': match[1], 'title':
                match[2]})

        return [header[0], result]

    else:
        return [header[0], None]


def read(category, disc_id, server_url=default_server,
         user=default_user, host=hostname, client_name=name,
         client_version=version):
    url = "%s?cmd=cddb+read+%s+%s&hello=%s+%s+%s+%s&proto=%i" % \
          (server_url, category, disc_id, user, host, client_name,
           client_version, proto)

    response = urllib.request.urlopen(url)

    header = response.readline().decode().rstrip().split(' ', 3)

    header[0] = int(header[0])
    if header[0] == 210 or header[0] == 417:  # success or access denied
        reply = []

        for line in response.readlines():
            line = line.decode().rstrip()

            if line == '.':
                break

            line = line.replace(r'\t', "\t")
            line = line.replace(r'\n', "\n")
            line = line.replace(r'\\', "\\")

            reply.append(line)

        if header[0] == 210:  # success, parse the reply
            return [header[0], parse_read_reply(reply)]
        else:  # access denied. :(
            return [header[0], reply]
    else:
        return [header[0], None]


def parse_read_reply(comments):
    len_re = re.compile(r'#\s*Disc length:\s*(\d+)\s*seconds')
    revis_re = re.compile(r'#\s*Revision:\s*(\d+)')
    submit_re = re.compile(r'#\s*Submitted via:\s*(.+)')
    keyword_re = re.compile(r'([^=]+)=(.*)')

    result = {}

    for line in comments:
        keyword_match = keyword_re.match(line)
        if keyword_match:
            (keyword, data) = keyword_match.groups()

            if keyword in result:
                result[keyword] = result[keyword] + data
            else:
                result[keyword] = data
            continue

        len_match = len_re.match(line)
        if len_match:
            result['disc_len'] = int(len_match.group(1))
            continue

        revis_match = revis_re.match(line)
        if revis_match:
            result['revision'] = int(revis_match.group(1))
            continue

        submit_match = submit_re.match(line)
        if submit_match:
            result['submitted_via'] = submit_match.group(1)
            continue

    return result



import logging
import os
import platform
import sys
import urllib.parse
from functools import partial

from PyQt5.QtCore import QDir, QSettings, QUrl, pyqtRemoveInputHook, pyqtSignal
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtWidgets import QAction, QApplication, QFileDialog, QFrame, QLabel, QMainWindow, QMenu, QMessageBox, QSplitter, QVBoxLayout, QWidget

from . import loadshortcuts as ls
from . import m3u, genres
from . import mainwin
from . import tagmodel
from . import version_string, changeset
from .mainwin import funcs as mainfuncs
from .masstag import dialogs
from .puddleobjects import (PuddleConfig, PuddleDock, winsettings,
                            progress, PuddleStatus, errormsg, dircmp, get_icon)
from .puddlesettings import SettingsDialog, load_gen_settings, update_settings
from .tagmodel import TagTable
from .about import versions

from . import audioinfo
from .util import rename_error_msg
from .audioinfo import lnglength, strlength, PATH, str_filesize, encode_fn
from errno import EEXIST
from operator import itemgetter
from collections import defaultdict
from . import constants, shortcutsettings

from . import findfunc, tagsources, confirmations
from . import action_shortcuts
import traceback
from . import plugins
from .translations import translate

status = PuddleStatus()

pyqtRemoveInputHook()

# A global variable that hold the status of
# various puddletag statuses.
# It is passed to any and all modules that asks for it.
# Feel free to read as much as you want from it, but
# modify only values that you've created or that are
# intended to be modified. This rule may be enforced
# in the future.
mainfuncs.status = status
tagmodel.status = status
mainwin.previews.set_status(status)
mainwin.tagtools.set_status(status)
plugins.status = status

confirmations.add('preview_mode', True, translate("Confirmations", 'Confirm when exiting preview mode.'))
confirmations.add('delete_files', True, translate("Confirmations", 'Confirm when deleting files.'))


def create_tool_windows(parent, extra=None):
    """Creates the dock widgets for the main window (parent) using
    the modules stored in puddlestuff/mainwin.

    Returns (the toggleViewActions of the docks, the dockWidgets the
    mselves)."""
    actions = []
    docks = []
    cparser = PuddleConfig()
    cparser.filename = ls.menu_path
    widgets = (mainwin.tagpanel, mainwin.artwork,
               mainwin.dirview, mainwin.patterncombo, mainwin.filterwin,
               mainwin.tagsources, mainwin.storedtags, mainwin.logdialog,
               dialogs)

    controls = [z.control for z in widgets]
    controls.extend(mainwin.action_dialogs.controls)
    if extra:
        controls.extend(extra)
    for z in controls:
        name = z[0]
        try:
            if not z[2]:
                PuddleDock._controls[name] = z[1](status=status)
                continue
        except IndexError:
            pass

        p = PuddleDock(z[0], z[1], parent, status)

        parent.addDockWidget(z[2], p)

        try:
            if z[4]: p.setFloating(True)
            p.move(parent.rect().center())
        except IndexError:
            pass

        p.setVisible(z[3])
        docks.append(p)
        action = p.toggleViewAction()
        action.setText(name)
        scut = cparser.get('winshortcuts', name, '')
        if scut:
            action.setShortcut(scut)
        actions.append(action)
    return actions, docks


def create_context_menus(controls, actions):
    """Creates context menus for controls using actions
    depending on whether and action in actions is
    supposed to be in a menu. See the context_menu
    function in the loadshortcuts module for more info."""
    for name in controls:
        menu = ls.context_menu(name, actions)
        if menu:
            controls[name].contextMenu = menu


def connect_controls(controls):
    """Connects the signals emitted by controls to any
    controls that receive them."""
    emits = {}
    for c in controls:
        for sig in c.emits:
            emits[sig].append(c) if sig in emits else emits.update({sig: [c]})

    for c in controls:
        for signal, slot in c.receives:
            if signal in emits:
                [getattr(i, signal).connect(slot) for i in emits[signal]]


def connect_control(control, controls):
    emits = defaultdict(lambda: [])
    for c in controls:
        [emits[sig].append(c) for sig in c.emits]

    for signal, slot in control.receives:
        if signal in emits:
            [getattr(c, signal).connect(slot) for c in emits[signal]]

    for c in controls:
        for signal, slot in c.receives:
            if signal in control.emits:
                getattr(control, signal).connect(slot)


def action_triggered_slot(control, command):
    "QAction.triggerred slot adds a checked=False param which we want to ignore"
    return lambda: getattr(control, command)()


def connect_actions(actions, controls):
    """Connect the triggered() signals in actions to the respective
    slot in controls if it exists. Just a message is shown if it
    doesn't."""
    emits = {}
    for c in controls.values():
        for sig in c.emits:
            emits[sig].append(c) if sig in emits else emits.update({sig: [c]})
    for action in actions:
        if action.enabled in emits:
            [getattr(c, action.enabled).connect(action.setEnabled)
             for c in emits[action.enabled]]
        else:
            logging.debug('No enable signal found for ' + action.text())
            action.setEnabled(False)
            continue
        if action.togglecheck and action.togglecheck in emits:
            [getattr(c, action.togglecheck).connect(action.setEnabled)
             for c in emits[action.togglecheck]]
        command = action.command
        if action.control == 'mainwin' and hasattr(mainfuncs, command):
            f = getattr(mainfuncs, command)
            if 'parent' in f.__code__.co_varnames:
                f = partial(f, parent=c)
            if action.togglecheck:
                action.toggled.connect(f)
            else:
                action.triggered.connect(f)
            continue
        elif action.control in controls:
            c = controls[action.control]
            if hasattr(c, command):
                action.triggered.connect(action_triggered_slot(c, command))
            else:
                logging.debug(action.command + ' slot not found for ' + action.text())


def connect_action_shortcuts(actions):
    cparser = PuddleConfig()
    cparser.filename = ls.menu_path
    for action in actions:
        shortcut = cparser.get('shortcuts', str(action.text()), '')
        if shortcut:
            action.setShortcut(shortcut)


def get_os():
    """Returns a string describing the underlying operating system."""

    def osrelease_dict_to_str(osrel):
        """Returns a pretty string from the given os-release dict."""
        # First try: The PRETTY_NAME usually already includes a version
        if 'PRETTY_NAME' in osrel:
            return osrel['PRETTY_NAME']

        # Second try: Combine name and version
        name = next((osrel[k] for k in ['NAME', 'ID'] if k in osrel), 'Linux')
        version = next((osrel[k] for k in ['VERSION', 'VERSION_ID', 'VERSION_CODENAME'] if k in osrel), '')
        return '%s %s' % (name, version)

    def read_osrelease():
        """Naive re-implementation of freedesktop_os_release."""
        for filepath in ['/etc/os-release', '/usr/lib/os-release']:
            osrelease_dict = {}
            try:
                with open(filepath, 'r', -1, 'utf-8') as file:
                    for line in file:
                        line = line.strip()
                        if line.startswith('#'):
                            continue
                        name, value = line.split('=', 1)
                        osrelease_dict[name.strip()] = value.strip('"\'').strip()
            except OSError:
                pass
            else:
                return osrelease_dict
        raise OSError('No file to read found')

    def get_linux_name():
        """Returns the name of the linux distribution. Hopefully."""
        # First try: The optional distro package (should always work if available)
        try:
            import distro
        except ImportError:
            pass
        else:
            return distro.name(True)

        # Second try: the new function in platform (py v3.10+ only)
        if hasattr(platform, 'freedesktop_os_release'):
            try:
                return osrelease_dict_to_str(platform.freedesktop_os_release())
            except OSError:
                pass

        # Third try: the old function in platform (until py v3.7 only)
        if hasattr(platform, 'linux_distribution'):
            linux_dist = platform.linux_distribution()
            if any((x for x in linux_dist if x)):
                return '%s %s' % (linux_dist[0], linux_dist[1])

        # Fourth try: read the os-release file directly (to fill the gap btw. 3.7 and 3.10)
        try:
            return osrelease_dict_to_str(read_osrelease())
        except OSError:
            pass

        # We tried hard, but it wasn't enough.
        return None

    if 'Linux' == platform.system():
        result = get_linux_name()
        if result:
            return result
        # else fall through to the fallback
    if 'Windows' == platform.system():
        win_info = platform.win32_ver()
        return 'Windows %s' % (win_info[0])
    if 'Darwin' == platform.system():
        return 'macOS %s' % (platform.mac_ver()[0])

    # In case of emergency, hope for the best
    return '%s %s %s' % (platform.system(), platform.release(), platform.version())


def create_bug_report_issue():
    """Create a new, pre-filled bug report as github issue."""
    puddletag_version = version_string
    if changeset:
        puddletag_version = f'{puddletag_version} ({changeset})'
    version_list = [
        ('Puddletag', puddletag_version),
        ('OS', get_os().strip()),
        *versions().items()
    ]

    params = urllib.parse.urlencode({
        'template': 'bug_report.yaml',
        'labels': 'bug',
        'system': '\n'.join(map(lambda x: f'{x[0]}: {x[1]}', version_list)),
    })
    url = f'https://github.com/puddletag/puddletag/issues/new?{params}'
    QDesktopServices.openUrl(QUrl(url))


def help_menu(parent):
    menu = QMenu(translate("Menus", 'Help'), parent)
    open_url = lambda url: QDesktopServices.openUrl(QUrl(url))

    connect = lambda c, s: c.triggered.connect(s)

    doc_link = QAction(translate("Menus", 'Online &Documentation'),
                       parent)
    connect(doc_link, lambda: open_url('https://docs.puddletag.net/docs.html'))

    forum_link = QAction(translate("Menus", '&GitHub project'), parent)
    connect(forum_link,
            lambda: open_url('https://github.com/puddletag/puddletag'))

    issue_link = QAction(translate("Menus", '&Report a problem'), parent)
    connect(issue_link, create_bug_report_issue)

    about_icon = get_icon('help-about')
    about = QAction(about_icon,
                    translate("Menus", 'About puddletag'), parent)
    connect(about, partial(mainfuncs.show_about, parent))

    about_qt = QAction(translate("Menus", 'About Qt'), parent)
    connect(about_qt, QApplication.aboutQt)

    sep = QAction(parent)
    sep.setSeparator(True)
    list(map(menu.addAction, (doc_link, forum_link, issue_link, sep,
                              about, about_qt)))

    return menu


def load_plugins():
    from .pluginloader import load_plugins
    from . import functions, musiclib
    plugins = load_plugins()
    findfunc.functions.update(plugins[constants.FUNCTIONS])
    functions.no_preview.extend(plugins[constants.FUNCTIONS_NO_PREVIEW])
    tagsources.tagsources.extend(plugins[constants.TAGSOURCE])
    musiclib.extralibs = plugins[constants.MUSICLIBS]

    return plugins[constants.DIALOGS], plugins[constants.MODULES]


class PreviewLabel(QLabel):
    valueChanged = pyqtSignal(bool, name='valueChanged')

    def __init__(self, *args, **kwargs):
        super(PreviewLabel, self).__init__(*args, **kwargs)
        self._enabled = False

    def mouseDoubleClickEvent(self, event):
        self._enabled = not self._enabled
        self.valueChanged.emit(self._enabled)


def _openFilesFilterFilename(filename):
    filename = os.path.abspath(filename)
    if isinstance(filename, str):
        filename = encode_fn(filename)
    return filename


# set in MainWin.__init__
add_shortcuts = None
remove_shortcuts = None


class MainWin(QMainWindow):
    loadFiles = pyqtSignal(object, object, object, object, object, name='loadFiles')
    always = pyqtSignal(bool, name='always')
    dirsmoved = pyqtSignal(list, name='dirsmoved')
    libfilesedited = pyqtSignal(list, name='libfilesedited')
    enable_preview_mode = pyqtSignal(name='enable_preview_mode')
    disable_preview_mode = pyqtSignal(name='disable_preview_mode')
    filesloaded = pyqtSignal(bool, name='filesloaded')
    filesselected = pyqtSignal(bool, name='filesselected')
    viewfilled = pyqtSignal(bool, name='viewfilled')

    def __init__(self):
        QMainWindow.__init__(self)

        self.__updateDirs = True
        self.__dirsToUpdate = []

        global add_shortcuts
        global remove_shortcuts
        add_shortcuts = self.addShortcuts
        remove_shortcuts = self.removeShortcuts
        plugins.add_shortcuts = add_shortcuts

        self.emits = ['loadFiles', 'always', 'dirsmoved', 'libfilesedited',
                      'enable_preview_mode', 'disable_preview_mode']

        self.receives = [('writeselected', self.writeTags),
                         ('filesloaded', self._filesLoaded),
                         ('viewfilled', self._viewFilled),
                         ('filesselected', self._filesSelected),
                         ('renamedirs', self.renameDirs),
                         ('filesloaded', self.updateTotalStats),
                         ('filesselected', self.updateSelectedStats),
                         ('onetomany', self.writeOneToMany),
                         ('dirschanged', self._dirChanged),
                         ('writepreview', self._writePreview),
                         ('clearpreview', self._clearPreview),
                         ('renameselected', self._renameSelected),
                         ('playlistchanged', self._dirChanged),
                         ('adddock', self.addDock),
                         ('writeaction', self.writeAction),
                         ('onetomanypreview', self.writeSinglePreview),
                         ('manypreview', self.writeManyPreview)]
        self.gensettings = [('&Load last folder at startup', False, 1)]
        self._playlist = None
        plugin_dialogs, plugin_modules = load_plugins()

        self.setWindowTitle("puddletag")
        self.setDockNestingEnabled(True)
        self._table = TagTable()
        self._table.dirsmoved.connect(self.updateDirs)
        win = QSplitter()

        layout = QVBoxLayout()
        layout.addWidget(self._table)
        layoutWidget = QWidget()
        layoutWidget.setLayout(layout)
        win.addWidget(layoutWidget)
        self.setCentralWidget(win)

        PuddleDock._controls = {
            'table': self._table,
            'mainwin': self,
            'funcs': mainfuncs.obj, }
        status['mainwin'] = self
        status['model'] = self._table.model()
        status['table'] = self._table

        ls.create_files()
        winactions, self._docks = create_tool_windows(self)
        status['dialogs'] = PuddleDock._controls
        self.createStatusBar()

        actions = ls.get_actions(self)
        menus = ls.get_menus('menu')
        previewactions = mainwin.previews.create_actions(self)

        all_actions = actions + winactions + previewactions

        controls = PuddleDock._controls

        toolgroup = ls.get_menus('toolbar')
        toolbar = ls.toolbar(toolgroup, all_actions, controls)
        toolbar.setObjectName(translate("Menus", 'Toolbar'))
        self.addToolBar(toolbar)

        menubar, winmenu, self._menus = ls.menubar(menus, all_actions)

        temp_winactions = winmenu.actions()
        [winmenu.addAction(a) for a in winactions if a not in temp_winactions]

        if winmenu:
            winmenu.addSeparator()
            self._winmenu = winmenu
        else:
            self._winmenu = QMenu(translate("Settings", '&Windows'), self)
            menubar.addMenu(self._winmenu)
        self.setMenuBar(menubar)
        menubar.addMenu(help_menu(self))
        mainfuncs.connect_status(actions)

        connect_actions(actions, controls)
        connect_action_shortcuts(all_actions)
        create_context_menus(controls, all_actions)
        status['actions'] = all_actions

        for m in plugin_modules:
            if hasattr(m, 'init'):
                try:
                    m.init(parent=self)
                except:

                    traceback.print_exc()
                    continue

        for win in plugin_dialogs:
            try:
                self.addDock(*win, connect=False)
            except:
                logging.exception("Error while loading Plugin dialog.")

        self.restoreSettings()
        self.always.emit(True)

    def addDock(self, name, dialog, position, visibility=True, connect=True):
        controls = list(PuddleDock._controls.values())
        dock = PuddleDock(name, dialog, self, status)
        self.addDockWidget(position, dock)
        self._winmenu.addAction(dock.toggleViewAction())
        if connect:
            connect_control(PuddleDock._controls[name], controls)
        dock.setVisible(visibility)
        self.restoreDockWidget(dock)
        return PuddleDock._controls[name]

    def addShortcuts(self, menu_title, actions, toolbar=False, save=False):
        if not actions:
            return

        if menu_title in self._menus:
            menu = self._menus[menu_title][0]
        else:
            menu = QMenu(menu_title)
            self._menus[menu_title] = [menu] + actions
            self.menuBar().insertMenu(self._menus['&Windows'][0].menuAction(), menu)

        status['actions'].extend(actions)
        list(map(menu.addAction, actions))

        if toolbar:
            list(map(self.toolbar.addAction, actions))
        if save:
            shortcutsettings.ActionEditorDialog.saveSettings(status['actions'])

    def _clearPreview(self):
        self._table.model().unSetTestData()

    def createShortcut(self, text, slot, *args, **kwargs):
        action = ls.create_action(self, text, None, slot)
        connect_actions([action], PuddleDock._controls)

    def _dirChanged(self, dirs):
        if not dirs:
            self.setWindowTitle('puddletag')
            return

        if isinstance(dirs, str):
            dirs = [dirs]

        dirs = [encode_fn(d) for d in dirs]

        if self._lastdir:
            initial = self._lastdir[0]
        else:
            initial = None

        if initial not in dirs:
            initial = dirs[0]

        if isinstance(initial, bytes):
            initial = initial.decode('utf8', 'replace')

        if len(dirs) > 1:
            self.setWindowTitle(translate('Main Window', "puddletag: {} + others").format(initial))
        else:
            self.setWindowTitle(translate('Main Window', "puddletag: {}").format(initial))

        self._lastdir = dirs

    def _getDir(self):
        dirname = self._lastdir[0] if self._lastdir else QDir.homePath()
        filedlg = QFileDialog()
        filedlg.setFileMode(QFileDialog.FileMode.Directory)
        filedlg.setOption(QFileDialog.Option.ShowDirsOnly)
        filename = str(QFileDialog.getExistingDirectory(self,
                                                        translate("Main Window", 'Import directory...'),
                                                        dirname,
                                                        QFileDialog.Option.ShowDirsOnly | QFileDialog.Option.DontUseNativeDialog | QFileDialog.Option.DontResolveSymlinks))
        return filename

    def appendDir(self, filename=None):
        self.openDir(filename, append=True)

    def _filesLoaded(self, val):
        self.filesloaded.emit(val)

    def _filesSelected(self, val):
        self.filesselected.emit(val)

    def applyGenSettings(self, settings, level=None):
        pass

    def closeEvent(self, e):
        preview_msg = translate('Previews',
                                'Some files have uncommited previews. '
                                'These changes will be lost once you exit puddletag. <br />'
                                'Do you want to exit without writing those changes?<br />')
        if tagmodel.has_previews(parent=self, msg=preview_msg):
            e.ignore()
            return False
        controls = PuddleDock._controls
        for control in PuddleDock._controls.values():
            if hasattr(control, 'saveSettings'):
                try:
                    control.saveSettings(self)
                except TypeError:
                    control.saveSettings()

        cparser = PuddleConfig()
        settings = QSettings(constants.QT_CONFIG, QSettings.Format.IniFormat)
        if self._lastdir:
            cparser.set('main', 'lastfolder', self._lastdir[0])
        cparser.set("main", "maximized", self.isMaximized())
        settings.setValue('main/state', self.saveState())

        headstate = self._table.horizontalHeader().saveState()
        settings.setValue('table/header', headstate)
        genres.save_genres(status['genres'])
        e.accept()

    def createStatusBar(self):
        statusbar = self.statusBar()
        statuslabel = QLabel()
        statuslabel.setFrameStyle(QFrame.Shape.NoFrame)
        statusbar.addPermanentWidget(statuslabel, 1)
        self._totalstats = QLabel('00 (00:00:00 | 00 MB)')
        self._selectedstats = QLabel('00 (00:00:00 | 00 MB)')
        preview_status = PreviewLabel(translate("Previews", 'Preview Mode: Off'))
        statusbar.addPermanentWidget(preview_status, 0)
        statusbar.addPermanentWidget(self._selectedstats, 0)
        statusbar.addPermanentWidget(self._totalstats, 0)

        def set_preview_status(value):
            if value:
                preview_status.setText(translate("Previews", '<b>Preview Mode: On</b>'))
            else:
                preview_status.setText(translate("Previews", 'Preview Mode: Off'))

        def change_preview(value):
            if value:
                self.enable_preview_mode.emit()
            else:
                self.disable_preview_mode.emit()

        preview_status.valueChanged.connect(change_preview)
        self._table.model().previewModeChanged.connect(set_preview_status)
        statusbar.setMaximumHeight(statusbar.height())
        statusbar.messageChanged.connect(statuslabel.setText)

    def loadPlayList(self):
        dirname = self._lastdir[0] if self._lastdir else QDir.homePath()
        selectedFile = QFileDialog.getOpenFileName(self,
                                                   translate("Playlist", translate("Playlist", 'Select m3u file...')),)
        filename = selectedFile[0]
        if not filename:
            return
        try:
            files = m3u.readm3u(filename)
            self.loadFiles.emit(files, None, None, None, filename)
        except (OSError, IOError) as e:
            QMessageBox.information(self._table,
                                    translate("Defaults", 'Error'),
                                    translate('Playlist',
                                              "An error occured while reading <b>{}</b> ({})"
                                              ).format(filename, e.strerror)
                                    )
        except (UnicodeError) as e:
            QMessageBox.information(self._table,
                                    translate("Defaults", 'Error'),
                                    translate("Playlist", 'The playlist is not encoded in UTF-8'))
        except Exception as e:
            QMessageBox.information(self._table, translate("Defaults", 'Error'),
                                    translate('Playlist',
                                              "An error occured while reading <b>{}</b> ({})"
                                              ).format(filename, str(e))
                                    )

    def openDir(self, filename=None, append=False):
        """Opens a folder. If filename != None, then
        the table is filled with the folder.

        If filename is None, show the open folder dialog and open that.

        If appenddir = True, the folder is appended.
        Otherwise, the folder is just loaded."""

        if filename is None:
            filename = self._getDir()
            if not filename:
                return
        else:
            if not isinstance(filename, str):
                filename = filename[0]

            filename = os.path.abspath(filename)

            if isinstance(filename, str):
                filename = encode_fn(filename)

        self.loadFiles.emit(None, [filename], append, None, None)

    def openFiles(self, filenames, append=False):
        filenames = map(_openFilesFilterFilename, filenames)
        self.loadFiles.emit(None, filenames, append, None, None)

    def openPrefs(self):
        win = SettingsDialog(list(PuddleDock._controls.values()), self, status)
        win.show()

    def removeShortcuts(self, menu_title, actions):
        if menu_title in self._menus:
            menu = self._menus[menu_title][0]
        if actions:
            children = dict([(str(z.text()), z) for z in menu.actions()])
            for action in actions:
                if isinstance(action, str):
                    action = children[action]
                menu.removeAction(action)
                try:
                    status['actions'].remove(action)
                except ValueError:
                    pass

    def restoreSettings(self):
        scts = action_shortcuts.create_action_shortcuts(
            mainwin.funcs.applyaction, self)

        self.addShortcuts('&Actions', scts)

        connect_actions(scts, PuddleDock._controls)

        cparser = PuddleConfig()
        settings = QSettings(constants.QT_CONFIG, QSettings.Format.IniFormat)

        gensettings = {}
        controls = list(PuddleDock._controls.values())
        for control in controls:
            if hasattr(control, 'loadSettings'):
                control.loadSettings()
            if hasattr(control, 'gensettings'):
                t = load_gen_settings(control.gensettings)
                gensettings[control] = dict(t)

        for control, val in gensettings.items():
            control.applyGenSettings(val, 0)

        self._lastdir = [encode_fn(cparser.get(
            'main', 'lastfolder', constants.HOMEDIR))]

        mapping = {
            'VorbisComment':
                {'date': 'year',
                 'tracknumber': 'track',
                 'musicbrainz_albumid': 'mbrainz_album_id',
                 'musicbrainz_artistid': 'mbrainz_artist_id',
                 'musicbrainz_trackid': 'mbrainz_track_id'},
            'MP4':
                {'MusicBrainz Track Id': 'mbrainz_track_id',
                 'MusicBrainz Artist Id': 'mbrainz_artist_id',
                 'MusicBrainz Album Id': 'mbrainz_album_id'},
            'ID3':
                {'ufid:http://musicbrainz.org': 'mbrainz_track_id',
                 'MusicBrainz Album Id': 'mbrainz_album_id',
                 'MusicBrainz Artist Id': 'mbrainz_artist_id'},
            'APEv2':
                {'musicbrainz_albumid': 'mbrainz_album_id',
                 'musicbrainz_artistid': 'mbrainz_artist_id',
                 'musicbrainz_trackid': 'mbrainz_track_id'}}

        filepath = os.path.join(cparser.savedir, 'mappings')
        audioinfo.setmapping(audioinfo.loadmapping(filepath, mapping))
        status['genres'] = genres.load_genres()

        connect_controls(controls + [mainwin.previews.obj])

        cover_pattern = cparser.get('tags', 'cover_pattern', 'folder')
        status['cover_pattern'] = cover_pattern

        winsettings('mainwin', self)
        if cparser.get("main", "maximized", True):
            self.showMaximized()

        QApplication.processEvents()

        if constants.FS_ENC == "ascii":
            QMessageBox.warning(self, "puddletag", translate("Errors",
                                                             "Your filesystem encoding was detected as <b>ASCII</b>. <br />"
                                                             "You won't be able to rename files using accented, <br />"
                                                             " cyrillic or any characters outside the ASCII alphabet."))

        for control, val in gensettings.items():
            control.applyGenSettings(val, 1)

        h = self._table.horizontalHeader()
        if settings.value('table/header'):
            h.restoreState(settings.value('table/header'))
        if settings.value('main/state'):
            self.restoreState(settings.value('main/state'))

        confirmations.load()
        shortcutsettings.ActionEditorDialog._loadSettings(status['actions'])
        update_settings()

        QApplication.processEvents()

    def savePlayList(self):
        tags = status['selectedfiles']
        if not tags:
            tags = status['alltags']
        settings = PuddleConfig()
        try:
            dirname = self._lastdir[0]
        except IndexError:
            dirname = constants.HOMEDIR
        filepattern = settings.get('playlist', 'filepattern', 'puddletag.m3u')
        default = encode_fn(findfunc.tagtofilename(filepattern, tags[0]))
        selectedFile = QFileDialog.getSaveFileName(self,
                                                   translate("Playlist", 'Save Playlist...'), os.path.join(dirname, default))
        f = selectedFile[0]
        if f:
            if settings.get('playlist', 'extinfo', 1, True):
                pattern = settings.get('playlist', 'extpattern', '%artist% - %title%')
            else:
                pattern = None

            reldir = settings.get('playlist', 'reldir', 0, True)
            windows_separator = settings.get('playlist', 'windows_separator', 0, False)
            m3u.exportm3u(tags, f, pattern, reldir, windows_separator)

    def _viewFilled(self, val):
        self.viewfilled.emit(val)

    def _updateStatus(self, files):
        if not files:
            return '00 (00:00:00 | 00 KB)'
        numfiles = len(files)
        stats = [(int(z.size), lnglength(z.length)) for z in files]
        totalsize = sum([z[0] for z in stats])
        totallength = strlength(sum([z[1] for z in stats]))

        sizetext = str_filesize(totalsize)
        return '%d (%s | %s)' % (numfiles, totallength, sizetext)

    def lockLayout(self):
        for dw in self._docks:
            dw.setTitleBarWidget(QWidget())

    def updateSelectedStats(self, *args):
        self._selectedstats.setText(self._updateStatus(status['selectedfiles']))

    def updateTotalStats(self, *args):
        self._totalstats.setText(self._updateStatus(status['alltags']))

    def _write(self, tagiter, rows=None, previews=None):
        self.__updateDirs = False
        if not rows:
            rows = status['selectedrows']
        model = self._table.model()
        setRowData = model.setRowData

        def fin():
            model.undolevel += 1
            self._table.selectionChanged()
            if not model.previewMode:
                self.libfilesedited.emit(lib_updates)

        lib_updates = []

        failed_rows = [rows[0]]  # First element=last row used.
        # Rest, rows that failed to write.

        if model.previewMode:
            [setRowData(row, f, undo=True) for row, f in zip(rows, tagiter)]
            fin()
            return

        def func():
            for row, f in zip(rows, tagiter):
                failed_rows[0] = row
                try:
                    update = setRowData(row, f, undo=True)
                    if update:
                        lib_updates.append(update)
                    yield None
                except PermissionError as e:
                    failed_rows.append(row)
                    filename = model.taginfo[row][PATH]
                    m = rename_error_msg(e, filename)
                    if row == rows[-1]:
                        yield m, 1
                    else:
                        yield m, len(rows)
                except EnvironmentError as e:
                    failed_rows.append(row)
                    filename = model.taginfo[row][PATH]
                    m = rename_error_msg(e, filename)
                    if row == rows[-1]:
                        yield m, 1
                    else:
                        yield m, len(rows)

        def finished():
            self.__updateDirs = True
            self.updateDirs([])
            if previews and failed_rows[1:]:
                model.previewMode = True
                last_row = failed_rows[0]
                [failed_rows.append(r) for r in previews if r > last_row]
                taginfo = model.taginfo
                for row in failed_rows[1:]:
                    if row not in previews:
                        continue
                    taginfo[row].preview = previews[row]
                last_row = failed_rows[0]

                model.updateTable(failed_rows)
            return fin()

        return func, finished, rows

    def writeTags(self, tagiter, rows=None, previews=None):
        ret = self._write(tagiter, rows, previews)
        if ret is None:
            return

        func, fin, rows = ret
        s = progress(func, translate("Defaults", 'Writing '), len(rows), fin)
        s(self)

    def writeAction(self, tagiter, rows=None, state=None):
        if state is None:
            state = {}
        ret = self._write(tagiter, rows)
        if ret is None:
            return
        func, fin, rows = ret

        def finished():
            fin()
            if 'rename_dirs' in state:
                self.renameDirs(list(state['rename_dirs'].items()))

        s = progress(func, translate("Defaults", 'Writing '), len(rows),
                     finished)
        s(self)

    def writeOneToMany(self, d):
        rows = status['selectedrows']
        ret = self._write((d.copy() for r in rows), rows)
        if ret is None:
            return
        func, fin, rows = ret

        s = progress(func, translate("Defaults", 'Writing '),
                     len(rows), fin)
        s(self)

    def writeSinglePreview(self, d):
        if not status['previewmode']:
            return
        model = self._table.model()
        rows = status['selectedrows']
        if not rows:
            return
        setRowData = model.setRowData

        [setRowData(row, d, undo=False, temp=True) for row in rows]
        columns = set(idx for idx, field in model.columns.items() if field in d)
        if columns:
            start = model.index(min(rows), min(columns))
            end = model.index(max(rows), max(columns))
            model.dataChanged.emit(start, end)

    def writeManyPreview(self, tags):
        if not status['previewmode']:
            return
        model = self._table.model()
        rows = status['selectedrows']
        setRowData = model.setRowData

        [setRowData(row, d, undo=False, temp=True) for row, d in
         zip(rows, tags)]
        columns = set(idx for idx, field in model.columns.items()
                      for tag in tags  for tagname in tag
                      if field == tagname)
        if columns:
            start = model.index(min(rows), min(columns))
            end = model.index(max(rows), max(columns))
            model.dataChanged.emit(start, end)

    def _writePreview(self):
        taginfo = self._table.model().taginfo
        previews = {}

        def get(audio, row):
            preview = audio.preview
            audio.preview = {}
            previews[row] = preview
            return row, preview

        data = [get(audio, row) for row, audio in
                enumerate(taginfo) if audio.preview]
        if not data:
            return
        self._table.model().previewMode = False
        self.writeTags((z[1] for z in data), [z[0] for z in data], previews)

    def _renameSelected(self, filenames):
        rows = status['selectedrows']
        files = status['selectedfiles']
        model = self._table.model()
        setRowData = model.setRowData

        def fin():
            model.undolevel += 1
            self._table.selectionChanged()

        if model.previewMode:
            for row, audio, filename in zip(rows, files, filenames):
                tag = PATH
                if tag in audio.mapping:
                    tag = audio.mapping[tag]
                setRowData(row, {tag: filename}, True, True)
            fin()
            return

        def func():
            for row, audio, filename in zip(rows, files, filenames):
                tag = PATH
                if tag in audio.mapping:
                    tag = audio.mapping[tag]
                try:
                    setRowData(row, {tag: filename}, True, True)
                    yield None
                except EnvironmentError as e:
                    m = translate('Dir Renaming',
                                  "An error occured while renaming <b>{}</b> to <b>{}</b>. ({})"
                                  ).format(audio[PATH], filename, e.strerror)
                    if row == rows[-1]:
                        yield m, 1
                    else:
                        yield m, len(rows)

        s = progress(func, translate("Dir Renaming", 'Renaming '), len(rows), fin)
        s(self)

    def renameDirs(self, dirs):
        self._table.saveSelection()
        showmessage = True
        dirs = sorted(dirs, dircmp, itemgetter(0))
        for index, (olddir, newdir) in enumerate(dirs):
            try:
                if os.path.exists(newdir) and (olddir != newdir):
                    raise IOError(EEXIST, os.strerror(EEXIST), newdir)
                os.rename(olddir, newdir)
                self._table.changeFolder(olddir, newdir)
                if self._lastdir and olddir in self._lastdir:
                    self._lastdir[self._lastdir.index(olddir)] = newdir
            except (IOError, OSError) as detail:
                msg = translate('Dir Renaming',
                                "I couldn't rename: <i>{}</i> to <b>{}</b> ({})"
                                ).format(olddir, newdir, detail.strerror)
                if index == len(dirs) - 1:
                    dirlen = 1
                else:
                    dirlen = len(dirs)
                if showmessage:
                    ret = errormsg(self, msg, dirlen)
                    if ret is True:
                        showmessage = False
                    elif ret is False:
                        break
        self.dirsmoved.emit(dirs)
        self._dirChanged(self._lastdir)
        self._table.restoreSelection()

    def updateDirs(self, dirs):
        if self.__updateDirs:
            if self.__dirsToUpdate:
                dirs = self.__dirsToUpdate + dirs
            self.__dirsToUpdate = []
        else:
            self.__dirsToUpdate.extend(dirs)
            return

        old_dirs = set()
        new_dirs = []
        for old_dir, new_dir in dirs:
            if old_dir not in old_dirs:
                new_dirs.append([old_dir, new_dir])
                old_dirs.add(old_dir)

        dirs = new_dirs

        if self._lastdir:
            last = self._lastdir[::]
        else:
            last = None
        for index, (olddir, newdir) in enumerate(dirs):
            self._table.changeFolder(olddir, newdir, False)
            if last and olddir in last:
                last[last.index(olddir)] = newdir
        self.dirsmoved.emit(dirs)
        self._dirChanged(last)


if __name__ == '__main__':
    app = QApplication(sys.argv)
    win = MainWin()
    win.show()
    app.exec_()



from copy import deepcopy

from mutagen.oggvorbis import OggVorbis

try:
    from mutagen.oggopus import OggOpus
except ImportError:
    OggOpus = None

from mutagen.flac import Picture, FLAC
import base64

from . import util
from .util import (usertags, PATH,
                   FILENAME, isempty, getdeco, setdeco, str_filesize, unicode_list,
                   CaselessDict, del_deco, keys_deco, fn_hash, cover_info,
                   get_total, set_total, parse_image, info_to_dict,
                   get_mime)
from .tag_versions import tags_in_file

PICARGS = ('type', 'mime', 'desc', 'width', 'height', 'depth', 'data')

ATTRIBUTES = ('frequency', 'length', 'bitrate', 'accessed', 'size', 'created',
              'modified', 'bitspersample')

COVER_KEY = 'metadata_block_picture'


def base64_to_image(value):
    return bin_to_image(Picture(base64.standard_b64decode(value)))


def bin_to_image(pic):
    return {'data': pic.data, 'description': pic.desc, 'mime': pic.mime,
            'imagetype': pic.type}


def image_to_base64(image):
    return base64.standard_b64encode(image_to_bin(image).write()).decode('ascii')


def image_to_bin(image):
    props = {}
    data = image[util.DATA]
    description = image.get(util.DESCRIPTION)
    if not description: description = ''

    mime = image.get(util.MIMETYPE)
    if mime is None: mime = get_mime(data)
    imagetype = image.get(util.IMAGETYPE, 3)

    props['type'] = imagetype
    props['desc'] = description
    props['mime'] = mime
    props['data'] = data

    p = Picture()
    [setattr(p, z, props[z]) for z in PICARGS if z in props]
    return p


def vorbis_tag(base, name):
    class Tag(util.MockTag):
        IMAGETAGS = (util.MIMETYPE, util.DESCRIPTION, util.DATA, util.IMAGETYPE)
        mapping = {}
        revmapping = {}

        def __init__(self, filename=None):
            self.__images = []
            self.__tags = CaselessDict()

            self.filetype = name
            self.__tags['__filetype'] = self.filetype
            self.__tags['__tag_read'] = 'VorbisComment'

            util.MockTag.__init__(self, filename)

        @property
        def filepath(self):
            return util.MockTag.get_filepath(self)

        @filepath.setter
        def filepath(self, val):
            self.__tags.update(util.MockTag.set_filepath(self, val))

        @property
        def images(self):
            return self.__images

        @images.setter
        def images(self, images):
            if images:
                self.__images = [parse_image(i, self.IMAGETAGS) for i in images]
            else:
                self.__images = []
            cover_info(images, self.__tags)

        def __contains__(self, key):
            if key == '__image':
                return bool(self.images)

            elif key == '__total':
                try:
                    return bool(get_total(self))
                except (KeyError, ValueError):
                    return False

            if self.revmapping:
                key = self.revmapping.get(key, key)
            return key in self.__tags

        def __deepcopy__(self, memo):
            cls = Tag()
            cls.mapping = self.mapping
            cls.revmapping = self.revmapping
            cls.set_fundamentals(deepcopy(self.__tags),
                                 self.mut_obj, self.images)
            cls.filepath = self.filepath
            return cls

        @del_deco
        def __delitem__(self, key):
            if key == '__image':
                self.images = []
            elif key.startswith('__'):
                return
            else:
                del (self.__tags[key])

        @getdeco
        def __getitem__(self, key):
            if key == '__image':
                return self.images
            elif key == '__total':
                return get_total(self)
            return self.__tags[key]

        @setdeco
        def __setitem__(self, key, value):
            if key.startswith('__'):
                if key == '__image':
                    self.images = value
                elif key == '__total':
                    set_total(self, value)
                elif key in fn_hash:
                    setattr(self, fn_hash[key], value)
            elif isempty(value):
                if key in self:
                    del (self[key])
                else:
                    return
            else:
                if isinstance(value, int):
                    self.__tags[key.lower()] = [str(value)]
                else:
                    self.__tags[key.lower()] = unicode_list(value)

        def delete(self):
            self.mut_obj.delete()
            for key in self.usertags:
                del (self.__tags[self.revmapping.get(key, key)])
            self.images = []

        @keys_deco
        def keys(self):
            return list(self.__tags.keys())

        def link(self, filename):
            """Links the audio, filename
            returns self if successful, None otherwise."""
            self.__images = []
            tags, audio = self.load(filename, base)
            if audio is None:
                return

            if audio.tags is not None and audio.tags.vendor:
                tags['__vendorstring'] = audio.tags.vendor
            for key in audio:
                if key == COVER_KEY:
                    self.images = list(map(base64_to_image, audio[key]))
                else:
                    self.__tags[key.lower()] = audio.tags[key]

            if base == FLAC:
                self.images = [_f for _f in map(bin_to_image, audio.pictures) if _f]
            elif not self.images:
                self.images = []

            self.__tags.update(info_to_dict(audio.info))
            self.__tags.update(tags)
            self.set_attrs(ATTRIBUTES)
            self.mut_obj = audio
            self._originaltags = list(self.__tags.keys())
            self.update_tag_list()
            return self

        def save(self):
            """Writes the tags in self.__tags
            to self.filename if no filename is specified."""
            filepath = self.filepath

            if self.mut_obj.tags is None:
                self.mut_obj.add_tags()
            if filepath != self.mut_obj.filename:
                self.mut_obj.filename = filepath
            audio = self.mut_obj

            newtag = {}
            for tag, value in usertags(self.__tags).items():
                newtag[tag] = value

            if self.__images:
                if base == FLAC:
                    audio.clear_pictures()
                    list(map(lambda p: audio.add_picture(image_to_bin(p)),
                             self.__images))
                else:
                    newtag[COVER_KEY] = [_f for _f in map(image_to_base64, self.__images) if _f]
            else:
                if base == FLAC:
                    audio.clear_pictures()

            toremove = [z for z in audio if z not in newtag]
            for z in toremove:
                del (audio[z])
            audio.update(newtag)
            audio.save()

        def set_fundamentals(self, tags, mut_obj, images=None):
            self.__tags = tags
            self.mut_obj = mut_obj
            if images:
                self.images = images
            else:
                self.images = []
            self._originaltags = list(tags.keys())

        def update_tag_list(self):
            l = tags_in_file(self.filepath)
            if l:
                self.__tags['__tag'] = 'VorbisComment, ' + ', '.join(l)
            else:
                self.__tags['__tag'] = 'VorbisComment'

    return Tag


class Ogg_Tag(vorbis_tag(OggVorbis, 'Ogg Vorbis')):
    @property
    def info(self):
        info = self.mut_obj.info
        fileinfo = [
            ('Path', self[PATH]),
            ('Size', str_filesize(int(self.size))),
            ('Filename', self[FILENAME]),
            ('Modified', self.modified)]

        ogginfo = [('Bitrate', self.bitrate),
                   ('Frequency', self.frequency),
                   ('Channels', str(info.channels)),
                   ('Length', self.length)]
        return [('File', fileinfo), ('Ogg Info', ogginfo)]


if OggOpus:
    class Opus_Tag(vorbis_tag(OggOpus, 'Ogg Opus')):
        @property
        def info(self):
            info = self.mut_obj.info
            fileinfo = [
                ('Path', self[PATH]),
                ('Size', str_filesize(int(self.size))),
                ('Filename', self[FILENAME]),
                ('Modified', self.modified)]

            ogginfo = [('Bitrate', self.bitrate),
                       ('Channels', str(info.channels)),
                       ('Length', self.length)]
            return [('File', fileinfo), ('Opus Info', ogginfo)]


class FLAC_Tag(vorbis_tag(FLAC, 'FLAC')):
    @property
    def info(self):
        info = self.mut_obj.info
        fileinfo = [('Path', self[PATH]),
                    ('Size', str_filesize(int(self.size))),
                    ('Filename', self[FILENAME]),
                    ('Modified', self.modified)]

        ogginfo = [('Bitrate', 'Lossless'),
                   ('Frequency', self.frequency),
                   ('Bits Per Sample', self.bitspersample),
                   ('Channels', str(info.channels)),
                   ('Length', self.length)]
        return [('File', fileinfo), ('FLAC Info', ogginfo)]


filetypes = []

if OggOpus:
    filetypes.append((OggOpus, Opus_Tag, 'VorbisComment', 'opus.ogg'))
    filetypes.append((OggOpus, Opus_Tag, 'VorbisComment', 'opus'))

filetypes.extend([
    (OggVorbis, Ogg_Tag, 'VorbisComment', 'ogg'),
    (FLAC, FLAC_Tag, 'VorbisComment', 'flac')]
)
